r"""jc - JSON Convert `route print` command output parser

See also: the `route` command parser

Usage (cli):

    $ route print | jc --route-print

Usage (module):

    import jc
    result = jc.parse('route_print', route_print_command_output)

Schema:

    {
        "interface_list": [
            {
                "interface_index":            integer,
                "mac_address":                string,
                "description":                string
            }
        ],
        "ipv4_route_table": {
            "active_routes": [
                {
                    "network_destination":    string,
                    "netmask":                string,
                    "gateway":                string,
                    "interface":              string,
                    "metric":                 integer,  # [0]
                    "metric_set_to_default":  boolean   # [1]
                }
            ],
            "persistent_routes": [
                {
                    "network_address":        string,
                    "netmask":                string,
                    "gateway_address":        string,
                    "metric":                 integer   # [0]
                    "metric_set_to_default":  boolean   # [1]
                }
            ]
        },
        "ipv6_route_table": {
            "active_routes": [
                {
                    "interface":              integer,
                    "metric":                 integer,  # [0]
                    "metric_set_to_default":  boolean,  # [1]
                    "network_destination":    string,
                    "gateway":                string
                }
            ],
            "persistent_routes": [
                {
                    "interface":              integer,
                    "metric":                 integer,  # [0]
                    "metric_set_to_default":  boolean,  # [1]
                    "network_destination":    string,
                    "gateway":                string
                }
            ]
        }
    }

    [0] Null/None if "metric" = "Default"
    [1] True if "metric" = "Default"

Examples:

    $ route print | jc --route-print -p
    {
      "interface_list": [
        {
          "interface_index": 28,
          "mac_address": null,
          "description": "Tailscale Tunnel"
        },
        {
          "interface_index": 12,
          "mac_address": "00:1c:42:da:01:6a",
          "description": "Parallels VirtIO Ethernet Adapter"
        },
        {
          "interface_index": 1,
          "mac_address": null,
          "description": "Software Loopback Interface 1"
        }
      ],
      "ipv4_route_table": {
        "active_routes": [
          {
            "network_destination": "0.0.0.0",
            "netmask": "0.0.0.0",
            "gateway": "10.211.55.1",
            "interface": "10.211.55.3",
            "metric": 15,
            "metric_set_to_default": false
          },
          {
            "network_destination": "10.0.0.0",
            "netmask": "255.0.0.0",
            "gateway": "192.168.22.1",
            "interface": "10.211.55.3",
            "metric": 16,
            "metric_set_to_default": false
          },
          ...
          {
            "network_destination": "255.255.255.255",
            "netmask": "255.255.255.255",
            "gateway": "On-link",
            "interface": "10.211.55.3",
            "metric": null,
            "metric_set_to_default": true
          }
        ],
        "persistent_routes": [
          {
            "network_address": "10.0.1.0",
            "netmask": "255.255.255.0",
            "gateway_address": "192.168.22.1",
            "metric": 1,
            "metric_set_to_default": false
          },
          {
            "network_address": "10.0.3.0",
            "netmask": "255.255.255.0",
            "gateway_address": "192.168.22.1",
            "metric": 1,
            "metric_set_to_default": false
          },
          ...
        ]
      },
      "ipv6_route_table": {
        "active_routes": [
          {
            "interface": 1,
            "metric": 331,
            "network_destination": "::1/128",
            "gateway": "On-link",
            "metric_set_to_default": false
          },
          {
            "interface": 12,
            "metric": 271,
            "network_destination": "2001:db8::/64",
            "gateway": "fe80::1",
            "metric_set_to_default": false
          },
          ...
          {
            "interface": 12,
            "metric": 271,
            "network_destination": "ff00::/8",
            "gateway": "On-link",
            "metric_set_to_default": false
          }
        ],
        "persistent_routes": [
          {
            "interface": 0,
            "metric": 4294967295,
            "network_destination": "2001:db8::/64",
            "gateway": "fe80::1",
            "metric_set_to_default": false
          }
        ]
      }
    }
"""
import re
import jc.utils


class info():
    """Provides parser metadata (version, author, etc.)"""
    version = '1.0'
    description = '`route print` command parser'
    author = 'joehacksalot'
    author_email = 'joehacksalot@gmail.com'
    details = 'See also: `route` command parser'
    compatible = ['win32']
    magic_commands = ['route print']
    tags = ['command']


__version__ = info.version


def parse(data, raw=False, quiet=False):
    """
    Main text parsing function

    Parameters:

        data:        (string)  text data to parse
        raw:         (boolean) unprocessed output if True
        quiet:       (boolean) suppress warning messages if True

    Returns:

        Parsed dictionary. The raw and processed data structures are the same.
    """
    jc.utils.compatibility(__name__, info.compatible, quiet)
    jc.utils.input_type_check(data)

    raw_output = {}
    if jc.utils.has_data(data):
        raw_data = {
                "interface_list": [],
                "ipv4_route_table": {
                    "active_routes": [],
                    "persistent_routes": []
                },
                "ipv6_route_table": {
                    "active_routes": [],
                    "persistent_routes": []
                }
            }

        lines = data.splitlines()
        _parse_interface_list(raw_data, _PushbackIterator(iter(lines)))
        _parse_ipv4_route_table(raw_data, _PushbackIterator(iter(lines)))
        _parse_ipv6_route_table(raw_data, _PushbackIterator(iter(lines)))
        raw_output = raw_data
    return raw_output if raw else _process(raw_output)

def _process(proc_data):
    """
    Final processing to conform to the schema.

    Parameters:

        proc_data:   (Dictionary) raw structured data to process

    Returns:

        Processed Dictionary. Structured data to conform to the schema.
    """
    if not proc_data:
        return {}

    for interface in proc_data['interface_list']:
        if interface["mac_address"] == '' or interface["mac_address"] == '00 00 00 00 00 00 00 e0':  # Placeholder MAC address for virtual adapters
            mac_address = None
        else:
            mac_address = interface["mac_address"].replace(" ", ":")

        interface["mac_address"] = mac_address

        interface["interface_index"] = jc.utils.convert_to_int(interface["interface_index"])

    for ipv4_active_route in proc_data['ipv4_route_table']['active_routes']:
        if ipv4_active_route["metric"] != "Default":
            ipv4_active_route["metric"] = jc.utils.convert_to_int(ipv4_active_route["metric"])
            ipv4_active_route["metric_set_to_default"] = False
        else:
            ipv4_active_route["metric"] = None
            ipv4_active_route["metric_set_to_default"] = True

    for ipv4_persistent_route in proc_data['ipv4_route_table']['persistent_routes']:
        if ipv4_persistent_route["metric"] != "Default":
            ipv4_persistent_route["metric"] = jc.utils.convert_to_int(ipv4_persistent_route["metric"])
            ipv4_persistent_route["metric_set_to_default"] = False
        else:
            ipv4_persistent_route["metric"] = None
            ipv4_persistent_route["metric_set_to_default"] = True

    for ipv6_active_route in proc_data['ipv6_route_table']['active_routes']:
        ipv6_active_route["interface"] = jc.utils.convert_to_int(ipv6_active_route["interface"])

        if ipv6_active_route["metric"] != "Default":
            ipv6_active_route["metric"] = jc.utils.convert_to_int(ipv6_active_route["metric"])
            ipv6_active_route["metric_set_to_default"] = False
        else:
            ipv6_active_route["metric"] = None
            ipv6_active_route["metric_set_to_default"] = True

    for ipv6_persistent_route in proc_data['ipv6_route_table']['persistent_routes']:
        ipv6_persistent_route["interface"] = jc.utils.convert_to_int(ipv6_persistent_route["interface"])

        if ipv6_persistent_route["metric"] != "Default":
            ipv6_persistent_route["metric"] = jc.utils.convert_to_int(ipv6_persistent_route["metric"])
            ipv6_persistent_route["metric_set_to_default"] = False
        else:
            ipv6_persistent_route["metric"] = None
            ipv6_persistent_route["metric_set_to_default"] = True

    return proc_data


class _PushbackIterator:
    """Iterator that allows pushing back values onto the iterator. Supports handing off
       parsing to localized parsers while maintaining line synchonization."""

    def __init__(self, iterator):
        self.iterator = iterator
        self.pushback_stack = []

    def __iter__(self):
        return self

    def __next__(self):
        if self.pushback_stack:
            return self.pushback_stack.pop()
        else:
            return next(self.iterator)

    def pushback(self, value):
        self.pushback_stack.append(value)

    def contains(self, pattern):
        iter_lines = list(self.iterator)
        list_lines = self.pushback_stack.copy()
        list_lines.extend(iter_lines)
        self.iterator = iter(list_lines)
        self.pushback_stack = []

        # Check the pushback stack first
        for line in list_lines:
            if re.match(pattern, line):
                return True
        return False

    def skip_until(self, pattern):
        for line in self:
            if re.match(pattern, line):
                return line
        return None


def _parse_interface_list(data, lines_iter):
    start_of_interface_list_pattern = r'^Interface List'
    if lines_iter.contains(start_of_interface_list_pattern):
        line = lines_iter.skip_until(start_of_interface_list_pattern)
        for line in lines_iter:
            if re.match(r'^=+$', line):
                break  # End of interface list
            interface_index = line[:5].replace(".", "").strip()
            mac_address = line[5:30].replace(".","").strip()
            description = line[30:].strip()
            data['interface_list'].append({
                "interface_index": interface_index,
                "mac_address": mac_address,
                "description": description
            })

def _parse_ipv4_route_table(data, lines_iter):
    def _parse_ipv4_active_routes(data, lines_iter):
        line = lines_iter.skip_until(r'^Active Routes')
        line = next(lines_iter, '') # Skip the header line
        if line.strip() == 'None':
            return
        for line in lines_iter:
            if re.match(r'^=+$', line):
                break  # End of interface list
            if 'Default Gateway' in line:
                continue
            lines_split = line.split()
            network_destination = lines_split[0]
            netmask = lines_split[1]
            gateway = lines_split[2]
            interface = lines_split[3]
            metric = lines_split[4]
            data['ipv4_route_table']["active_routes"].append({
                "network_destination": network_destination,
                "netmask": netmask,
                "gateway": gateway,
                "interface": interface,
                "metric": metric
            })

    def _parse_ipv4_persistent_routes(data, lines_iter):
        line = lines_iter.skip_until(r'^Persistent Routes')
        line = next(lines_iter, '') # line is either "None" and we abort parsing this section or we skip header line
        if line.strip() == 'None':
            return
        for line in lines_iter:
            if re.match(r'^=+$', line):
                break 
            lines_split = line.split()
            network_address = lines_split[0]
            netmask = lines_split[1]
            gateway_address = lines_split[2]
            metric = lines_split[3]
            data['ipv4_route_table']["persistent_routes"].append({
                "network_address": network_address,
                "netmask": netmask,
                "gateway_address": gateway_address,
                "metric": metric
            })

    start_of_ipv4_route_table_pattern = r'^IPv4 Route Table'
    if lines_iter.contains(start_of_ipv4_route_table_pattern):
        line = lines_iter.skip_until(start_of_ipv4_route_table_pattern)
        line = next(lines_iter, '') # Skip the separator line
        _parse_ipv4_active_routes(data, lines_iter)
        _parse_ipv4_persistent_routes(data, lines_iter)

def _parse_ipv6_route_table(data, lines_iter):
    def _parse_ipv6_active_routes(data, lines_iter):
        line = lines_iter.skip_until(r'^Active Routes')
        line = next(lines_iter, '') # line is either "None" and we abort parsing this section or we skip header line
        if line.strip() == 'None':
            return
        for line in lines_iter:
            if re.match(r'^=+$', line):
                break
            split_line = line.split()
            interface = split_line[0]
            metric = split_line[1]
            network_destination = split_line[2]
            if len(split_line) > 3:
                gateway = split_line[3]
            else:
                gateway = next(lines_iter, '').strip()
            data['ipv6_route_table']["active_routes"].append({
                "interface": interface,
                "metric": metric,
                "network_destination": network_destination,
                "gateway": gateway
            })

    def _parse_ipv6_persistent_routes(data, lines_iter):
        line = lines_iter.skip_until(r'^Persistent Routes')
        line = next(lines_iter, '') # line is either "None" and we abort parsing this section or we skip header line
        if line.strip() == 'None':
            return
        for line in lines_iter:
            if re.match(r'^=+$', line):
                break
            split_line = line.split()
            interface = split_line[0]
            metric = split_line[1]
            network_destination = split_line[2]
            if len(split_line) > 3:
                gateway = split_line[3]
            else:
                gateway = next(lines_iter, '').strip()
            data['ipv6_route_table']["persistent_routes"].append({
                "interface": interface,
                "metric": metric,
                "network_destination": network_destination,
                "gateway": gateway
            })

    start_of_ipv6_route_table_pattern = r'^IPv6 Route Table'
    if lines_iter.contains(start_of_ipv6_route_table_pattern):
        line = lines_iter.skip_until(start_of_ipv6_route_table_pattern)
        line = next(lines_iter, '') # Skip the separator line
        _parse_ipv6_active_routes(data, lines_iter)
        _parse_ipv6_persistent_routes(data, lines_iter)



import os
import json
import unittest
import jc.parsers.uname
from jc.exceptions import ParseError

THIS_DIR = os.path.dirname(os.path.abspath(__file__))


class MyTests(unittest.TestCase):

    # input
    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/uname-a.out'), 'r', encoding='utf-8') as f:
        centos_7_7_uname_a = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/uname.out'), 'r', encoding='utf-8') as f:
        centos_7_7_uname = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/ubuntu-18.04/uname-a.out'), 'r', encoding='utf-8') as f:
        ubuntu_18_4_uname_a = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/osx-10.11.6/uname-a.out'), 'r', encoding='utf-8') as f:
        osx_10_11_6_uname_a = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/osx-10.14.6/uname-a.out'), 'r', encoding='utf-8') as f:
        osx_10_14_6_uname_a = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/osx-10.14.6/uname.out'), 'r', encoding='utf-8') as f:
        osx_10_14_6_uname = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/freebsd12/uname-a.out'), 'r', encoding='utf-8') as f:
        freebsd12_uname_a = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/freebsd12/uname-a2.out'), 'r', encoding='utf-8') as f:
        freebsd12_uname_a2 = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/uname-a.out'), 'r', encoding='utf-8') as f:
        generic_uname_a = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/uname-a-different-proc.out'), 'r', encoding='utf-8') as f:
        generic_uname_a_different_proc = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/debian10/uname-a.out'), 'r', encoding='utf-8') as f:
        debian_10_uname_a = f.read()

    # output
    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/uname-a.json'), 'r', encoding='utf-8') as f:
        centos_7_7_uname_a_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/ubuntu-18.04/uname-a.json'), 'r', encoding='utf-8') as f:
        ubuntu_18_4_uname_a_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/osx-10.11.6/uname-a.json'), 'r', encoding='utf-8') as f:
        osx_10_11_6_uname_a_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/osx-10.14.6/uname-a.json'), 'r', encoding='utf-8') as f:
        osx_10_14_6_uname_a_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/freebsd12/uname-a.json'), 'r', encoding='utf-8') as f:
        freebsd12_uname_a_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/freebsd12/uname-a2.json'), 'r', encoding='utf-8') as f:
        freebsd12_uname_a2_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/uname-a.json'), 'r', encoding='utf-8') as f:
        generic_uname_a_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/uname-a-different-proc.json'), 'r', encoding='utf-8') as f:
        generic_uname_a_different_proc_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/debian10/uname-a.json'), 'r', encoding='utf-8') as f:
        debian_10_uname_a_json = json.loads(f.read())


    def test_uname_nodata(self):
        """
        Test 'uname -a' with no data
        """
        self.assertEqual(jc.parsers.uname.parse('', quiet=True), {})

    def test_uname_no_a_osx(self):
        """
        Test 'uname' without -a option on OSX. Should generate a ParseError exception
        """
        self.assertRaises(ParseError, jc.parsers.uname.parse, self.osx_10_14_6_uname, quiet=True)

    def test_uname_no_a_centos(self):
        """
        Test 'uname' without -a option on Centos. Should generate a ParseError exception
        """
        self.assertRaises(ParseError, jc.parsers.uname.parse, self.centos_7_7_uname, quiet=True)

    def test_uname_centos_7_7(self):
        """
        Test 'uname -a' on Centos 7.7
        """
        self.assertEqual(jc.parsers.uname.parse(self.centos_7_7_uname_a, quiet=True), self.centos_7_7_uname_a_json)

    def test_uname_ubuntu_18_4(self):
        """
        Test 'uname -a' on Ubuntu 18.4
        """
        self.assertEqual(jc.parsers.uname.parse(self.ubuntu_18_4_uname_a, quiet=True), self.ubuntu_18_4_uname_a_json)

    def test_uname_osx_10_11_6(self):
        """
        Test 'uname -a' on OSX 10.11.6
        """
        self.assertEqual(jc.parsers.uname.parse(self.osx_10_11_6_uname_a, quiet=True), self.osx_10_11_6_uname_a_json)

    def test_uname_osx_10_14_6(self):
        """
        Test 'uname -a' on OSX 10.14.6
        """
        self.assertEqual(jc.parsers.uname.parse(self.osx_10_14_6_uname_a, quiet=True), self.osx_10_14_6_uname_a_json)

    def test_uname_freebsd12(self):
        """
        Test 'uname -a' on freebsd12
        """
        self.assertEqual(jc.parsers.uname.parse(self.freebsd12_uname_a, quiet=True), self.freebsd12_uname_a_json)

    def test_uname2_freebsd12(self):
        """
        Test 'uname -a' on freebsd12 with longer version level string
        """
        self.assertEqual(jc.parsers.uname.parse(self.freebsd12_uname_a2, quiet=True), self.freebsd12_uname_a2_json)

    def test_uname_generic(self):
        """
        Test 'uname -a' on debian with missing hardware platform and processor
        """
        self.assertEqual(jc.parsers.uname.parse(self.generic_uname_a, quiet=True), self.generic_uname_a_json)

    def test_uname_different_proc_generic(self):
        """
        Test 'uname -a' on machine with different processor type
        """
        self.assertEqual(jc.parsers.uname.parse(self.generic_uname_a_different_proc, quiet=True), self.generic_uname_a_different_proc_json)

    def test_uname_debian_10(self):
        """
        Test 'uname -a' on debian 10 with missing hardware platform and processor
        """
        self.assertEqual(jc.parsers.uname.parse(self.debian_10_uname_a, quiet=True), self.debian_10_uname_a_json)


if __name__ == '__main__':
    unittest.main()



r"""jc - JSON Convert `last` and `lastb` command output parser

Supports `-w`, `-F`, and `-x` options.

Calculated epoch time fields are naive (i.e. based on the local time of the
system the parser is run on) since there is no timezone information in the
`last` command output.

Usage (cli):

    $ last | jc --last

or

    $ jc last

Usage (module):

    import jc
    result = jc.parse('last', last_command_output)

Schema:

    [
      {
        "user":             string,
        "tty":              string,
        "hostname":         string,
        "login":            string,
        "logout":           string,
        "duration":         string,
        "login_epoch":      integer,  # (naive) available w/last -F option
        "logout_epoch":     integer,  # (naive) available w/last -F option
        "duration_seconds": integer   # available w/last -F option
      }
    ]

Examples:

    $ last -F | jc --last -p
    [
      {
        "user": "kbrazil",
        "tty": "ttys002",
        "hostname": null,
        "login": "Mon Dec 28 17:24:10 2020",
        "logout": "still logged in"
      },
      {
        "user": "kbrazil",
        "tty": "ttys003",
        "hostname": null,
        "login": "Mon Dec 28 17:24:10 2020",
        "logout": "Mon Dec 28 17:25:01 2020",
        "duration": "00:00",
        "login_epoch": 1565891826,
        "logout_epoch": 1565895404,
        "duration_seconds": 3578
      },
      {
        "user": "kbrazil",
        "tty": "ttys003",
        "hostname": null,
        "login": "Mon Dec 28 17:24:10 2020",
        "logout": "Mon Dec 28 17:25:01 2020",
        "duration": "00:00",
        "login_epoch": 1565891826,
        "logout_epoch": 1565895404,
        "duration_seconds": 3578
      },
      ...
    ]

    $ last | jc --last -p -r
    [
      {
        "user": "kbrazil",
        "tty": "ttys002",
        "hostname": "-",
        "login": "Thu Feb 27 14:31",
        "logout": "still_logged_in"
      },
      {
        "user": "kbrazil",
        "tty": "ttys003",
        "hostname": "-",
        "login": "Thu Feb 27 10:38",
        "logout": "10:38",
        "duration": "00:00"
      },
      {
        "user": "kbrazil",
        "tty": "ttys003",
        "hostname": "-",
        "login": "Thu Feb 27 10:18",
        "logout": "10:18",
        "duration": "00:00"
      },
      ...
    ]
"""
import re
import jc.utils

DATE_RE = re.compile(r'[MTWFS][ouerha][nedritnu] [JFMASOND][aepuco][nbrynlgptvc]')
LAST_F_DATE_RE = re.compile(r'\d\d:\d\d:\d\d \d\d\d\d')
LOGIN_LOGOUT_EPOCH_RE = re.compile(r'.*\d\d:\d\d:\d\d \d\d\d\d.*')
LOGOUT_IGNORED_EVENTS = ['down', 'crash']


class info():
    """Provides parser metadata (version, author, etc.)"""
    version = '1.9'
    description = '`last` and `lastb` command parser'
    author = 'Kelly Brazil'
    author_email = 'kellyjonbrazil@gmail.com'
    details = 'Enhancements by https://github.com/zerolagtime'
    compatible = ['linux', 'darwin', 'aix', 'freebsd']
    magic_commands = ['last', 'lastb']
    tags = ['command']


__version__ = info.version


def _process(proc_data):
    """
    Final processing to conform to the schema.

    Parameters:

        proc_data:   (List of Dictionaries) raw structured data to process

    Returns:

        List of Dictionaries. Structured data to conform to the schema.
    """
    for entry in proc_data:
        if 'user' in entry and entry['user'] == 'boot_time':
            entry['user'] = 'boot time'

        if 'tty' in entry and entry['tty'] == '~':
            entry['tty'] = None

        if 'hostname' in entry and entry['hostname'] == '-':
            entry['hostname'] = None

        if 'hostname' in entry and entry['hostname'] is not None and entry['hostname'][0] == ":":
            entry['hostname'] = f'CONSOLE{entry["hostname"]}'

        if 'logout' in entry and entry['logout'] == 'still_logged_in':
            entry['logout'] = 'still logged in'

        if 'logout' in entry and entry['logout'] == 'gone_-_no_logout':
            entry['logout'] = 'gone - no logout'

        if 'login' in entry and LOGIN_LOGOUT_EPOCH_RE.match(entry['login']):
            timestamp = jc.utils.timestamp(entry['login'])
            entry['login_epoch'] = timestamp.naive

        if 'logout' in entry and LOGIN_LOGOUT_EPOCH_RE.match(entry['logout']):
            timestamp = jc.utils.timestamp(entry['logout'])
            entry['logout_epoch'] = timestamp.naive

        if 'login_epoch' in entry and 'logout_epoch' in entry:
            entry['duration_seconds'] = entry['logout_epoch'] - entry['login_epoch']

        if 'duration' in entry and re.match(r'^\d+\+', entry['duration']):
            m = re.match(r'^(?P<days>\d+)\+(?P<hours>\d\d):(?P<minutes>\d\d)', entry['duration'])
            days, hours, minutes = m.groups()
            entry['duration'] = f'{int(days)*24 + int(hours)}:{minutes}'

    return proc_data


def parse(data, raw=False, quiet=False):
    """
    Main text parsing function

    Parameters:

        data:        (string)  text data to parse
        raw:         (boolean) unprocessed output if True
        quiet:       (boolean) suppress warning messages if True

    Returns:

        List of Dictionaries. Raw or processed structured data.
    """
    jc.utils.compatibility(__name__, info.compatible, quiet)
    jc.utils.input_type_check(data)

    raw_output = []

    # Clear any blank lines
    cleandata = list(filter(None, data.splitlines()))

    if not jc.utils.has_data(data):
        return []

    for entry in cleandata:
        output_line = {}

        if any(
            entry.startswith(f'{prefix} begins ')
            for prefix in ['wtmp', 'btmp', 'utx.log']
        ):
            continue

        entry = entry.replace('boot time', 'boot_time')
        entry = entry.replace('  still logged in', '- still_logged_in')
        entry = entry.replace('  gone - no logout', '- gone_-_no_logout')

        linedata = entry.split()

        # Adding "-" before the date part.
        if DATE_RE.match(' '.join(linedata[2:4])):
            linedata.insert(2, '-')

        # freebsd fix
        if linedata[0] == 'boot_time':
            linedata.insert(1, '-')
            linedata.insert(1, '~')

        output_line['user'] = linedata[0]

        # Fix for last -x (runlevel).
        if output_line['user'] == 'runlevel' and  linedata[1] == '(to':
            linedata[1] += f' {linedata.pop(2)} {linedata.pop(2)}'
        elif output_line['user'] in ['reboot', 'shutdown'] and linedata[1] == 'system':  # system down\system boot
            linedata[1] += f' {linedata.pop(2)}'

        output_line['tty'] = linedata[1]
        output_line['hostname'] = linedata[2]

        # last -F support
        if LAST_F_DATE_RE.match(' '.join(linedata[6:8])):
            output_line['login'] = ' '.join(linedata[3:8])

            if len(linedata) > 9:
                if linedata[9] not in LOGOUT_IGNORED_EVENTS:
                    output_line['logout'] = ' '.join(linedata[9:14])
                else:
                    output_line['logout'] = linedata[9]
                    # add more items to the list to line up duration
                    for _ in range(4):
                        linedata.insert(10, '-')

            if len(linedata) > 14:
                output_line['duration'] = linedata[14].replace('(', '').replace(')', '')
        else: # normal last support
            output_line['login'] = ' '.join(linedata[3:7])

            if len(linedata) > 8:
                output_line['logout'] = linedata[8]

            if len(linedata) > 9:
                output_line['duration'] = linedata[9].replace('(', '').replace(')', '')

        raw_output.append(output_line)

    if raw:
        return raw_output

    return _process(raw_output)



r"""jc - JSON Convert `/proc/<pid>/mountinfo` file parser

Usage (cli):

    $ cat /proc/1/mountinfo | jc --proc

or

    $ jc /proc/1/mountinfo

or

    $ cat /proc/1/mountinfo | jc --proc-pid-mountinfo

Usage (module):

    import jc
    result = jc.parse('proc', proc_pid_mountinfo_file)

or

    import jc
    result = jc.parse('proc_pid_mountinfo', proc_pid_mountinfo_file)

Schema:

    [
      {
        "mount_id":                 integer,
        "parent_id":                integer,
        "maj":                      integer,
        "min":                      integer,
        "root":                     string,
        "mount_point":              string,
        "mount_options": [
                                    string
        ],
        "optional_fields": {                   # [0]
          "<key>":                  integer    # [1]
        },
        "fs_type":                  string,
        "mount_source":             string,
        "super_options": [
                                    integer    # [2]
        ],
        "super_options_fields": {
          "<key>":                  string
        }
      }
    ]

    [0] if empty, then private mount
    [1] unbindable will always have a value of 0
    [2] integer conversions are attempted. Use --raw or raw=True for
        original string values.

Examples:

    $ cat /proc/1/mountinfo | jc --proc -p
    [
      {
        "mount_id": 24,
        "parent_id": 30,
        "maj": 0,
        "min": 22,
        "root": "/",
        "mount_point": "/sys",
        "mount_options": [
          "rw",
          "nosuid",
          "nodev",
          "noexec",
          "relatime"
        ],
        "optional_fields": {
          "master": 1,
          "shared": 7
        },
        "fs_type": "sysfs",
        "mount_source": "sysfs",
        "super_options": [
          "rw"
        ]
      },
      {
        "mount_id": 25,
        "parent_id": 30,
        "maj": 0,
        "min": 23,
        "root": "/",
        "mount_point": "/proc",
        "mount_options": [
          "rw",
          "nosuid",
          "nodev",
          "noexec",
          "relatime"
        ],
        "optional_fields": {
          "shared": 14
        },
        "fs_type": "proc",
        "mount_source": "proc",
        "super_options": [
          "rw"
        ]
      },
      ...
    ]

    $ cat /proc/1/mountinfo | jc --proc-pid-mountinfo -p -r
    [
      {
        "mount_id": "24",
        "parent_id": "30",
        "maj": "0",
        "min": "22",
        "root": "/",
        "mount_point": "/sys",
        "mount_options": "rw,nosuid,nodev,noexec,relatime",
        "optional_fields": "master:1 shared:7 ",
        "fs_type": "sysfs",
        "mount_source": "sysfs",
        "super_options": "rw"
      },
      {
        "mount_id": "25",
        "parent_id": "30",
        "maj": "0",
        "min": "23",
        "root": "/",
        "mount_point": "/proc",
        "mount_options": "rw,nosuid,nodev,noexec,relatime",
        "optional_fields": "shared:14 ",
        "fs_type": "proc",
        "mount_source": "proc",
        "super_options": "rw"
      },
      ...
    ]
"""
import re
from typing import List, Dict
import jc.utils


class info():
    """Provides parser metadata (version, author, etc.)"""
    version = '1.0'
    description = '`/proc/<pid>/mountinfo` file parser'
    author = 'Kelly Brazil'
    author_email = 'kellyjonbrazil@gmail.com'
    compatible = ['linux']
    tags = ['file']
    hidden = True


__version__ = info.version


def _process(proc_data: List[Dict]) -> List[Dict]:
    """
    Final processing to conform to the schema.

    Parameters:

        proc_data:   (List of Dictionaries) raw structured data to process

    Returns:

        List of Dictionaries. Structured to conform to the schema.
    """
    int_list = {'mount_id', 'parent_id', 'maj', 'min'}

    for entry in proc_data:
        for key in entry:
            if key in int_list:
                entry[key] = int(entry[key])

        if 'mount_options' in entry:
            entry['mount_options'] = entry['mount_options'].split(',')

        if 'optional_fields' in entry:
            if 'unbindable' in  entry['optional_fields']:
                entry['optional_fields'] = {'unbindable': 0}
            else:
                entry['optional_fields'] = {x.split(':')[0]: int(x.split(':')[1]) for x in entry['optional_fields'].split()}

        if 'super_options' in entry:
            if entry['super_options']:
                super_options_split = entry['super_options'].split(',')
                s_options = [x for x in super_options_split if '=' not in x]
                s_options_fields = [x for x in super_options_split if '=' in x]

                if s_options:
                    entry['super_options'] = s_options
                else:
                    del entry['super_options']

                if s_options_fields:
                    if not 'super_options_fields' in entry:
                        entry['super_options_fields'] = {}

                    for field in s_options_fields:
                        key, val = field.split('=')
                        entry['super_options_fields'][key] = jc.utils.convert_to_int(val)

            else:
                del entry['super_options']

    return proc_data


def parse(
    data: str,
    raw: bool = False,
    quiet: bool = False
) -> List[Dict]:
    """
    Main text parsing function

    Parameters:

        data:        (string)  text data to parse
        raw:         (boolean) unprocessed output if True
        quiet:       (boolean) suppress warning messages if True

    Returns:

        List of Dictionaries. Raw or processed structured data.
    """
    jc.utils.compatibility(__name__, info.compatible, quiet)
    jc.utils.input_type_check(data)

    raw_output: List = []

    if jc.utils.has_data(data):

        line_pattern = re.compile(r'''
            ^(?P<mount_id>\d+)\s
            (?P<parent_id>\d+)\s
            (?P<maj>\d+):
            (?P<min>\d+)\s
            (?P<root>\S+)\s
            (?P<mount_point>\S+)\s
            (?P<mount_options>\S+)\s?
            # (?P<optional_fields>(?:\s?\S+:\S+\s?)*)\s?-\s
            (?P<optional_fields>(?:\s?(?:\S+:\S+|unbindable)\s?)*)\s?-\s
            (?P<fs_type>\S+)\s
            (?P<mount_source>\S+)\s
            (?P<super_options>\S+)?
            ''', re.VERBOSE
        )

        for line in filter(None, data.splitlines()):

            line_match = line_pattern.search(line)
            if line_match:
                raw_output.append(line_match.groupdict())

    return raw_output if raw else _process(raw_output)



r"""jc - JSON Convert `lsattr` command output parser

> Note: filenames with newlines are not supported.

Usage (cli):

    $ lsattr | jc --lsattr

or

    $ jc lsattr

Usage (module):

    import jc
    result = jc.parse('lsattr', lsattr_command_output)

Schema:

Information from https://github.com/mirror/busybox/blob/2d4a3d9e6c1493a9520b907e07a41aca90cdfd94/e2fsprogs/e2fs_lib.c#L40
used to define field names

    [
      {
        "file":                           string,
        "compressed_file":                Optional[boolean],
        "compressed_dirty_file":          Optional[boolean],
        "compression_raw_access":         Optional[boolean],
        "secure_deletion":                Optional[boolean],
        "undelete":                       Optional[boolean],
        "synchronous_updates":            Optional[boolean],
        "synchronous_directory_updates":  Optional[boolean],
        "immutable":                      Optional[boolean],
        "append_only":                    Optional[boolean],
        "no_dump":                        Optional[boolean],
        "no_atime":                       Optional[boolean],
        "compression_requested":          Optional[boolean],
        "encrypted":                      Optional[boolean],
        "journaled_data":                 Optional[boolean],
        "indexed_directory":              Optional[boolean],
        "no_tailmerging":                 Optional[boolean],
        "top_of_directory_hierarchies":   Optional[boolean],
        "extents":                        Optional[boolean],
        "no_cow":                         Optional[boolean],
        "casefold":                       Optional[boolean],
        "inline_data":                    Optional[boolean],
        "project_hierarchy":              Optional[boolean],
        "verity":                         Optional[boolean],
      }
    ]

Examples:

      $ sudo lsattr /etc/passwd | jc --lsattr
      [
        {
            "file": "/etc/passwd",
            "extents": true
        }
      ]
"""
import re
from typing import List, Dict
from jc.jc_types import JSONDictType
import jc.utils


class info():
    """Provides parser metadata (version, author, etc.)"""
    version = '1.1'
    description = '`lsattr` command parser'
    author = 'Mark Rotner'
    author_email = 'rotner.mr@gmail.com'
    compatible = ['linux']
    magic_commands = ['lsattr']
    tags = ['command']


__version__ = info.version


LINE_RE = re.compile(r'(?P<attributes>[BZXsuSDiadAcEjItTeCFNPV-]{20}) (?P<filename>.*)')

# https://github.com/mirror/busybox/blob/2d4a3d9e6c1493a9520b907e07a41aca90cdfd94/e2fsprogs/e2fs_lib.c#L40
# https://github.com/landley/toybox/blob/f1682dc79fd75f64042b5438918fe5a507977e1c/toys/other/lsattr.c#L97
ATTRIBUTES = {
    "B": "compressed_file",
    "Z": "compressed_dirty_file",
    "X": "compression_raw_access",
    "s": "secure_deletion",
    "u": "undelete",
    "S": "synchronous_updates",
    "D": "synchronous_directory_updates",
    "i": "immutable",
    "a": "append_only",
    "d": "no_dump",
    "A": "no_atime",
    "c": "compression_requested",
    "E": "encrypted",
    "j": "journaled_data",
    "I": "indexed_directory",
    "t": "no_tailmerging",
    "T": "top_of_directory_hierarchies",
    "e": "extents",
    "C": "no_cow",
    "F": "casefold",
    "N": "inline_data",
    "P": "project_hierarchy",
    "V": "verity",
}


def _process(proc_data: List[JSONDictType]) -> List[JSONDictType]:
    """
    Final processing to conform to the schema.

    Parameters:

        proc_data:   (List of Dictionaries) raw structured data to process

    Returns:

        List of Dictionaries. Structured to conform to the schema.
    """
    for item in proc_data:
        if 'attributes' in item:
            for attribute in item['attributes']:
                attribute_key = ATTRIBUTES.get(attribute)
                if attribute_key:
                    item[attribute_key] = True
            del item['attributes']
    return proc_data


def parse(
    data: str,
    raw: bool = False,
    quiet: bool = False
) -> List[JSONDictType]:
    """
    Main text parsing function

    Parameters:

        data:        (string)  text data to parse
        quiet:       (boolean) suppress warning messages if True

    Returns:

        List of Dictionaries. Raw or processed structured data.
    """
    jc.utils.compatibility(__name__, info.compatible, quiet)
    jc.utils.input_type_check(data)

    output: List = []

    cleandata = list(filter(None, data.splitlines()))

    if not jc.utils.has_data(data):
        return output

    for line in cleandata:
        # --------------e----- /etc/passwd
        line_split = re.match(LINE_RE, line)

        if line_split:
            attributes = line_split['attributes']
            file = line_split['filename']

            line_output: Dict = {}
            line_output['file'] = file
            line_output['attributes'] = attributes

            if line_output:
                output.append(line_output)

    return output if raw else _process(output)



r"""jc - JSON Convert `/proc/consoles` file parser

Usage (cli):

    $ cat /proc/consoles | jc --proc

or

    $ jc /proc/consoles

or

    $ cat /proc/consoles | jc --proc-consoles

Usage (module):

    import jc
    result = jc.parse('proc', proc_consoles_file)

or

    import jc
    result = jc.parse('proc_consoles', proc_consoles_file)

Schema:

    [
      {
        "device":                     string,
        "operations":                 string,
        "operations_list": [
                                      string  # [0]
        ],
        "flags":                      string,
        "flags_list": [
                                      string  # [1]
        ],
        "major":                      integer,
        "minor":                      integer
      }
    ]

    [0] Values: read, write, unblank
    [1] Values: enabled, preferred, primary boot, prink buffer,
                braille device, safe when CPU offline

Examples:

    $ cat /proc/consoles | jc --proc -p
    [
      {
        "device": "tty0",
        "operations": "-WU",
        "operations_list": [
          "write",
          "unblank"
        ],
        "flags": "ECp",
        "flags_list": [
          "enabled",
          "preferred",
          "printk buffer"
        ],
        "major": 4,
        "minor": 7
      },
      {
        "device": "ttyS0",
        "operations": "-W-",
        "operations_list": [
          "write"
        ],
        "flags": "Ep",
        "flags_list": [
          "enabled",
          "printk buffer"
        ],
        "major": 4,
        "minor": 64
      }
    ]
"""
import shlex
from typing import List, Dict
import jc.utils


class info():
    """Provides parser metadata (version, author, etc.)"""
    version = '1.0'
    description = '`/proc/consoles` file parser'
    author = 'Kelly Brazil'
    author_email = 'kellyjonbrazil@gmail.com'
    compatible = ['linux']
    tags = ['file']
    hidden = True


__version__ = info.version


def _process(proc_data: List[Dict]) -> List[Dict]:
    """
    Final processing to conform to the schema.

    Parameters:

        proc_data:   (List of Dictionaries) raw structured data to process

    Returns:

        List of Dictionaries. Structured to conform to the schema.
    """
    int_list = {'major', 'minor'}

    for entry in proc_data:
        for key in entry:
            if key in int_list:
                entry[key] = jc.utils.convert_to_int(entry[key])

    return proc_data


def parse(
    data: str,
    raw: bool = False,
    quiet: bool = False
) -> List[Dict]:
    """
    Main text parsing function

    Parameters:

        data:        (string)  text data to parse
        raw:         (boolean) unprocessed output if True
        quiet:       (boolean) suppress warning messages if True

    Returns:

        List of Dictionaries. Raw or processed structured data.
    """
    jc.utils.compatibility(__name__, info.compatible, quiet)
    jc.utils.input_type_check(data)

    raw_output: List = []

    operations_map = {
        'R': 'read',
        'W': 'write',
        'U': 'unblank'
    }

    flags_map = {
        'E': 'enabled',
        'C': 'preferred',
        'B': 'primary boot',
        'p': 'printk buffer',
        'b': 'braille device',
        'a': 'safe when CPU offline'
    }

    if jc.utils.has_data(data):

        for line in filter(None, data.splitlines()):

            # since parens are acting like quotation marks, use shlex.split()
            # after converting parens to quotes.
            line = line.replace('(', '"'). replace(')', '"')
            device, operations, flags, maj_min = shlex.split(line)

            operations_str = operations.replace('-', '')
            operations_list = [operations_map[i] for i in operations_str]

            flags_str = flags.replace (' ', '')
            flags_list = [flags_map[i] for i in flags_str]

            raw_output.append(
                {
                    'device': device,
                    'operations': operations,
                    'operations_list': operations_list,
                    'flags': flags,
                    'flags_list': flags_list,
                    'major': maj_min.split(':')[0],
                    'minor': maj_min.split(':')[1]
                }
            )

    return raw_output if raw else _process(raw_output)



r"""jc - JSON Convert `wg show` command output parser

Parses the output of the `wg show all dump` command, providing structured JSON output for easy integration and analysis.

Usage (cli):

    $ wg show all dump | jc --wg-show

or

    $ jc wg show all dump

Usage (module):

    import jc
    result = jc.parse('wg-show', wg_command_output)

Schema:

    [
      {
        "device":                       string,
        "private_key":                  string,
        "public_key":                   string,
        "listen_port":                  integer,
        "fwmark":                       integer,
        "peers": [
          {
            "public_key":               string,
            "preshared_key":            string,
            "endpoint":                 string,
            "latest_handshake":         integer,
            "transfer_rx":              integer,
            "transfer_sx":              integer,
            "persistent_keepalive":     integer,
            "allowed_ips": [
                                        string
            ]
          }
        ]
      }
    ]

Examples:

    $ wg show all dump | jc --wg-show -p
    [
      {
        "device": "wg0",
        "private_key": "aEbVdvHSEp3oofHDNVCsUoaRSxk1Og8/pTLof5yF+1M=",
        "public_key": "OIxbQszw1chdO5uigAxpsl4fc/h04yMYafl72gUbakM=",
        "listen_port": 51820,
        "fwmark": null,
        "peers": [
          {
            "public_key": "sQFGAhSdx0aC7DmTFojzBOW8Ccjv1XV5+N9FnkZu5zc=",
            "preshared_key": null,
            "endpoint": "79.134.136.199:40036",
            "latest_handshake": 1728809756,
            "transfer_rx": 1378724,
            "transfer_sx": 406524,
            "persistent_keepalive": null,
            "allowed_ips": ["10.10.0.2/32"]
          },
          {
            "public_key": "B9csmpvrv4Q7gpjc6zAbNNO8hIOYfpBqxmik2aNpwwE=",
            "preshared_key": null,
            "endpoint": "79.134.136.199:35946",
            "latest_handshake": 1728809756,
            "transfer_rx": 4884248,
            "transfer_sx": 3544596,
            "persistent_keepalive": null,
            "allowed_ips": ["10.10.0.3/32"]
          },
          {
            "public_key": "miiSYR5UdevREhlWpmnci+vv/dEGLHbNtKu7u1CuOD4=",
            "preshared_key": null,
            "allowed_ips": ["10.10.0.4/32"]
          },
          {
            "public_key": "gx9+JHLHJvOfBNjTmZ8KQAnThFFiZMQrX1kRaYcIYzw=",
            "preshared_key": null,
            "endpoint": "173.244.225.194:45014",
            "latest_handshake": 1728809827,
            "transfer_rx": 1363652,
            "transfer_sx": 458252,
            "persistent_keepalive": null,
            "allowed_ips": ["10.10.0.5/32"]
          }
        ]
      }
    ]

    $ wg show all dump | jc --wg-show -p -r
    [
      {
        "device": "wg0",
        "private_key": "aEbVdvHSEp3oofHDNVCsUoaRSxk1Og8/pTLof5yF+1M=",
        "public_key": "OIxbQszw1chdO5uigAxpsl4fc/h04yMYafl72gUbakM=",
        "listen_port": 51820,
        "fwmark": null,
        "peers": {
          "sQFGAhSdx0aC7DmTFojzBOW8Ccjv1XV5+N9FnkZu5zc=": {
            "preshared_key": null,
            "endpoint": "79.134.136.199:40036",
            "latest_handshake": 1728809756,
            "transfer_rx": 1378724,
            "transfer_sx": 406524,
            "persistent_keepalive": -1,
            "allowed_ips": ["10.10.0.2/32"]
          },
          "B9csmpvrv4Q7gpjc6zAbNNO8hIOYfpBqxmik2aNpwwE=": {
            "preshared_key": null,
            "endpoint": "79.134.136.199:35946",
            "latest_handshake": 1728809756,
            "transfer_rx": 4884248,
            "transfer_sx": 3544596,
            "persistent_keepalive": -1,
            "allowed_ips": ["10.10.0.3/32"]
          },
          "miiSYR5UdevREhlWpmnci+vv/dEGLHbNtKu7u1CuOD4=": {
            "preshared_key": null,
            "allowed_ips": ["10.10.0.4/32"]
          },
          "gx9+JHLHJvOfBNjTmZ8KQAnThFFiZMQrX1kRaYcIYzw=": {
            "preshared_key": null,
            "endpoint": "173.244.225.194:45014",
            "latest_handshake": 1728809827,
            "transfer_rx": 1363652,
            "transfer_sx": 458252,
            "persistent_keepalive": -1,
            "allowed_ips": ["10.10.0.5/32"]
          }
        }
      }
    ]
"""

from typing import List, Dict, Optional, Union
from jc.jc_types import JSONDictType
import jc.utils
import re

PeerData = Dict[str, Union[Optional[str], Optional[int], List[str]]]
DeviceData = Dict[str, Union[Optional[str], Optional[int], Dict[str, PeerData]]]


class info:
    """Provides parser metadata (version, author, etc.)"""

    version = "1.0"
    description = "`wg show` command parser"
    author = "Hamza Saht"
    author_email = "hamzasaht01@gmail.com"
    compatible = ["linux", "darwin", "cygwin", "win32", "aix", "freebsd"]
    tags = ["command"]
    magic_commands = ["wg show"]


__version__ = info.version


def _process(proc_data: List[DeviceData]) -> List[JSONDictType]:
    """
    Final processing to conform to the schema.

    Parameters:

        proc_data:   (List[Dict]) Raw structured data to process

    Returns:

        List[Dict]: Structured data that conforms to the schema
    """
    processed_data: List[JSONDictType] = []
    for device in proc_data:
        processed_device = {
            "device": device["device"],
            "private_key": device.get("private_key"),
            "public_key": device.get("public_key"),
            "listen_port": device.get("listen_port"),
            "fwmark": device.get("fwmark"),
            "peers": [
                {
                    "public_key": peer_key,
                    "preshared_key": peer_data.get("preshared_key"),
                    "endpoint": peer_data.get("endpoint"),
                    "latest_handshake": peer_data.get("latest_handshake", 0),
                    "transfer_rx": peer_data.get("transfer_rx", 0),
                    "transfer_sx": peer_data.get("transfer_sx", 0),
                    "persistent_keepalive": peer_data.get("persistent_keepalive", -1),
                    "allowed_ips": peer_data.get("allowed_ips", []),
                }
                for peer_key, peer_data in device.get("peers", {}).items()
            ],
        }
        processed_data.append(processed_device)
    return processed_data


def parse(data: str, raw: bool = False, quiet: bool = False) -> List[JSONDictType]:
    """
    Main text parsing function.

    Parses the output of the `wg` command, specifically `wg show all dump`, into structured JSON format.

    Parameters:

        data:        (str)  Text data to parse, typically the output from `wg show all dump`
        raw:         (bool) If True, returns unprocessed output
        quiet:       (bool) Suppress warning messages if True

    Returns:

        List[Dict]: Parsed data in JSON-friendly format, either raw or processed.
    """
    jc.utils.compatibility(__name__, info.compatible, quiet)
    jc.utils.input_type_check(data)

    raw_output: List[DeviceData] = []
    current_device: Optional[str] = None
    device_data: DeviceData = {}

    if jc.utils.has_data(data):
        for line in filter(None, data.splitlines()):
            fields = re.split(r"\s+", line.strip())
            if len(fields) == 5:
                device, private_key, public_key, listen_port, fwmark = fields
                if current_device:
                    raw_output.append({"device": current_device, **device_data})
                current_device = device
                device_data = {
                    "private_key": private_key if private_key != "(none)" else None,
                    "public_key": public_key if public_key != "(none)" else None,
                    "listen_port": int(listen_port) if listen_port != "0" else None,
                    "fwmark": int(fwmark) if fwmark != "off" else None,
                    "peers": {},
                }
            elif len(fields) == 9:
                (
                    interface,
                    public_key,
                    preshared_key,
                    endpoint,
                    allowed_ips,
                    latest_handshake,
                    transfer_rx,
                    transfer_tx,
                    persistent_keepalive,
                ) = fields
                peer_data: PeerData = {
                    "preshared_key": preshared_key
                    if preshared_key != "(none)"
                    else None,
                    "endpoint": endpoint if endpoint != "(none)" else None,
                    "latest_handshake": int(latest_handshake),
                    "transfer_rx": int(transfer_rx),
                    "transfer_sx": int(transfer_tx),
                    "persistent_keepalive": int(persistent_keepalive)
                    if persistent_keepalive != "off"
                    else -1,
                    "allowed_ips": allowed_ips.split(",")
                    if allowed_ips != "(none)"
                    else [],
                }
                device_data["peers"][public_key] = {
                    k: v for k, v in peer_data.items() if v is not None
                }

        if current_device:
            raw_output.append({"device": current_device, **device_data})

    return raw_output if raw else _process(raw_output)



r"""jc - JSON Convert `rsync` command output parser

Supports the `-i` or `--itemize-changes` options with all levels of
verbosity. This parser will process the `STDOUT` output or a log file
generated with the `--log-file` option.

The `--stats` or `--info=stats[1-3]` options are also supported.

Usage (cli):

    $ rsync -i -a source/ dest | jc --rsync

or

    $ jc rsync -i -a source/ dest

or

    $ cat rsync-backup.log | jc --rsync

Usage (module):

    import jc
    result = jc.parse('rsync', rsync_command_output)

Schema:

    [
      {
        "summary": {
          "date":                             string,
          "time":                             string,
          "process":                          integer,
          "sent":                             integer,
          "received":                         integer,
          "total_size":                       integer,
          "matches":                          integer,
          "hash_hits":                        integer,
          "false_alarms":                     integer,
          "data":                             integer,
          "bytes_sec":                        float,
          "speedup":                          float,
          "total_files":                      integer,
          "regular_files":                    integer,
          "dir_files":                        integer,
          "total_created_files":              integer,
          "created_regular_files":            integer,
          "created_dir_files":                integer,
          "deleted_files":                    integer,
          "transferred_files":                integer,
          "transferred_file_size":            integer,
          "literal_data":                     integer,
          "matched_data":                     integer,
          "file_list_size":                   integer,
          "file_list_generation_time":        float,
          "file_list_transfer_time":          float,
        },
        "files": [
          {
            "filename":                       string,
            "date":                           string,
            "time":                           string,
            "process":                        integer,
            "metadata":                       string,
            "update_type":                    string/null,  # [0]
            "file_type":                      string/null,  # [1]
            "checksum_or_value_different":    bool/null,
            "size_different":                 bool/null,
            "modification_time_different":    bool/null,
            "permissions_different":          bool/null,
            "owner_different":                bool/null,
            "group_different":                bool/null,
            "acl_different":                  bool/null,
            "extended_attribute_different":   bool/null,
            "epoch":                          integer,      # [2]
          }
        ]
      }
    ]

    Size values are in bytes.

    [0] 'file sent', 'file received', 'local change or creation',
        'hard link', 'not updated', 'message'
    [1] 'file', 'directory', 'symlink', 'device', 'special file'
    [2] naive timestamp if time and date fields exist and can be converted.

Examples:

    $ rsync -i -a source/ dest | jc --rsync -p
    [
      {
        "summary": {
          "sent": 1708,
          "received": 8209,
          "bytes_sec": 19834.0,
          "total_size": 235,
          "speedup": 0.02
        },
        "files": [
          {
            "filename": "./",
            "metadata": ".d..t......",
            "update_type": "not updated",
            "file_type": "directory",
            "checksum_or_value_different": false,
            "size_different": false,
            "modification_time_different": true,
            "permissions_different": false,
            "owner_different": false,
            "group_different": false,
            "acl_different": false,
            "extended_attribute_different": false
          },
          ...
        ]
      }
    ]

    $ rsync | jc --rsync -p -r
    [
      {
        "summary": {
          "sent": "1,708",
          "received": "8,209",
          "bytes_sec": "19,834.00",
          "total_size": "235",
          "speedup": "0.02"
        },
        "files": [
          {
            "filename": "./",
            "metadata": ".d..t......",
            "update_type": "not updated",
            "file_type": "directory",
            "checksum_or_value_different": false,
            "size_different": false,
            "modification_time_different": true,
            "permissions_different": false,
            "owner_different": false,
            "group_different": false,
            "acl_different": false,
            "extended_attribute_different": false
          },
          ...
        ]
      }
    ]
"""
import re
from copy import deepcopy
from typing import List, Dict, Union
import jc.utils


class info():
    """Provides parser metadata (version, author, etc.)"""
    version = '1.3'
    description = '`rsync` command parser'
    author = 'Kelly Brazil'
    author_email = 'kellyjonbrazil@gmail.com'
    compatible = ['linux', 'darwin', 'freebsd']
    magic_commands = ['rsync -i', 'rsync --itemize-changes']
    tags = ['command']


__version__ = info.version


def _process(proc_data: List[Dict]) -> List[Dict]:
    """
    Final processing to conform to the schema.

    Parameters:

        proc_data:   (List of Dictionaries) raw structured data to process

    Returns:

        List of Dictionaries. Structured to conform to the schema.
    """
    int_list = {
        'process', 'sent', 'received', 'total_size', 'matches', 'hash_hits',
        'false_alarms', 'data', 'total_files', 'regular_files', 'dir_files',
        'total_created_files', 'created_regular_files', 'created_dir_files',
        'deleted_files', 'transferred_files', 'transferred_file_size',
        'literal_data', 'matched_data', 'file_list_size'
    }

    float_list = {
        'bytes_sec', 'speedup', 'file_list_generation_time',
        'file_list_transfer_time'
    }

    for item in proc_data:
        for key in item['summary']:
            if key in int_list:
                item['summary'][key] = jc.utils.convert_size_to_int(item['summary'][key])

            if key in float_list:
                converted_val: Union[float, None] = None
                val = item['summary'][key]
                if any([
                    'K' in val,
                    'M' in val,
                    'G' in val,
                    'T' in val
                ]):
                    converted_int_val = jc.utils.convert_size_to_int(val)

                    if not converted_int_val is None:
                        converted_val = float(converted_int_val)

                else:
                    converted_val = jc.utils.convert_to_float(val)

                item['summary'][key] = converted_val

        for entry in item['files']:
            for key in entry:
                if key in int_list:
                    entry[key] = jc.utils.convert_to_int(entry[key])

            # add timestamp
            if 'date' in entry and 'time' in entry:
                date = entry['date'].replace('/', '-')
                date_time = f'{date} {entry["time"]}'
                ts = jc.utils.timestamp(date_time, format_hint=(7250,))
                entry['epoch'] = ts.naive

    return proc_data


def parse(
    data: str,
    raw: bool = False,
    quiet: bool = False
) -> List[Dict]:
    """
    Main text parsing function

    Parameters:

        data:        (string)  text data to parse
        raw:         (boolean) unprocessed output if True
        quiet:       (boolean) suppress warning messages if True

    Returns:

        List of Dictionaries. Raw or processed structured data.
    """
    jc.utils.compatibility(__name__, info.compatible, quiet)
    jc.utils.input_type_check(data)

    raw_output: List = []

    rsync_run_new: Dict = {
        'summary': {},
        'files': []
    }

    rsync_run = deepcopy(rsync_run_new)

    last_process = ''

    update_type = {
        '<': 'file sent',
        '>': 'file received',
        'c': 'local change or creation',
        'h': 'hard link',
        '.': 'not updated',
        '*': 'message',
        '+': None
    }

    file_type = {
        'f': 'file',
        'd': 'directory',
        'L': 'symlink',
        'D': 'device',
        'S': 'special file',
        '+': None
    }

    checksum_or_value_different = {
        'c': True,
        '.': False,
        '+': None,
        ' ': None,
        '?': None
    }

    size_different = {
        's': True,
        '.': False,
        '+': None,
        ' ': None,
        '?': None
    }

    modification_time_different = {
        't': True,
        '.': False,
        '+': None,
        ' ': None,
        '?': None
    }

    permissions_different = {
        'p': True,
        '.': False,
        '+': None,
        ' ': None,
        '?': None
    }

    owner_different = {
        'o': True,
        '.': False,
        '+': None,
        ' ': None,
        '?': None
    }

    group_different = {
        'g': True,
        '.': False,
        '+': None,
        ' ': None,
        '?': None
    }

    acl_different = {
        'a': True,
        '.': False,
        '+': None,
        ' ': None,
        '?': None
    }

    extended_attribute_different = {
        'x': True,
        '.': False,
        '+': None,
        ' ': None,
        '?': None
    }

    file_line_re = re.compile(r'(?P<meta>[<>ch.*][fdlDS][c.+ ?][s.+ ?][t.+ ?][p.+ ?][o.+ ?][g.+ ?][u.+ ?][a.+ ?][x.+ ?]) (?P<name>.+)')
    file_line_mac_re = re.compile(r'(?P<meta>[<>ch.*][fdlDS][c.+ ?][s.+ ?][t.+ ?][p.+ ?][o.+ ?][g.+ ?][x.+ ?]) (?P<name>.+)')
    stat1_line_re = re.compile(r'(sent)\s+(?P<sent>[0-9,]+)\s+(bytes)\s+(received)\s+(?P<received>[0-9,]+)\s+(bytes)\s+(?P<bytes_sec>[0-9,.]+)\s+(bytes/sec)')
    stat2_line_re = re.compile(r'(total size is)\s+(?P<total_size>[0-9,]+)\s+(speedup is)\s+(?P<speedup>[0-9,.]+)')

    stat1_line_simple_re = re.compile(r'(sent)\s+(?P<sent>[0-9,.TGMK]+)\s+(bytes)\s+(received)\s+(?P<received>[0-9,.TGMK]+)\s+(bytes)\s+(?P<bytes_sec>[0-9,.TGMK]+)\s+(bytes/sec)')
    stat2_line_simple_re = re.compile(r'(total\s+size\s+is)\s+(?P<total_size>[0-9,.TGMK]+)\s+(speedup\s+is)\s+(?P<speedup>[0-9,.TGMK]+)')

    file_line_log_re = re.compile(r'(?P<date>\d\d\d\d/\d\d/\d\d)\s+(?P<time>\d\d:\d\d:\d\d)\s+\[(?P<process>\d+)\]\s+(?P<meta>[<>ch.*][fdlDS][c.+ ?][s.+ ?][t.+ ?][p.+ ?][o.+ ?][g.+ ?][u.+ ?][a.+ ?][x.+ ?]) (?P<name>.+)')
    file_line_log_mac_re = re.compile(r'(?P<date>\d\d\d\d/\d\d/\d\d)\s+(?P<time>\d\d:\d\d:\d\d)\s+\[(?P<process>\d+)\]\s+(?P<meta>[<>ch.*][fdlDS][c.+ ?][s.+ ?][t.+ ?][p.+ ?][o.+ ?][g.+ ?][x.+ ?]) (?P<name>.+)')
    stat_line_log_re = re.compile(r'(?P<date>\d\d\d\d/\d\d/\d\d)\s+(?P<time>\d\d:\d\d:\d\d)\s+\[(?P<process>\d+)\]\s+sent\s+(?P<sent>[\d,]+)\s+bytes\s+received\s+(?P<received>[\d,]+)\s+bytes\s+total\s+size\s+(?P<total_size>[\d,]+)')

    stat1_line_log_v_re = re.compile(r'(?P<date>\d\d\d\d/\d\d/\d\d)\s+(?P<time>\d\d:\d\d:\d\d)\s+\[(?P<process>\d+)]\s+total:\s+matches=(?P<matches>[\d,]+)\s+hash_hits=(?P<hash_hits>[\d,]+)\s+false_alarms=(?P<false_alarms>[\d,]+)\s+data=(?P<data>[\d,]+)')
    stat2_line_log_v_re = re.compile(r'(?P<date>\d\d\d\d/\d\d/\d\d)\s+(?P<time>\d\d:\d\d:\d\d)\s+\[(?P<process>\d+)\]\s+sent\s+(?P<sent>[\d,]+)\s+bytes\s+received\s+(?P<received>[\d,]+)\s+bytes\s+(?P<bytes_sec>[\d,.]+)\s+bytes/sec')
    stat3_line_log_v_re = re.compile(r'(?P<date>\d\d\d\d/\d\d/\d\d)\s+(?P<time>\d\d:\d\d:\d\d)\s+\[(?P<process>\d+)]\s+total\s+size\s+is\s+(?P<total_size>[\d,]+)\s+speedup\s+is\s+(?P<speedup>[\d,.]+)')

    stat_ex_files_number_re = re.compile(r'Number\sof\sfiles:\s(?P<files_total>[,0123456789]+)\s\(reg:\s(?P<files_regular>[,0123456789]+),\sdir:\s(?P<files_dir>[,0123456789]+)\)$')
    stat_ex_files_created_re = re.compile(r'Number\sof\screated\sfiles:\s(?P<files_created_total>[,0123456789]+)\s\(reg:\s(?P<files_created_regular>[,0123456789]+),\sdir:\s(?P<files_created_dir>[,0123456789]+)\)$')
    stat_ex_files_deleted_re = re.compile(r'Number\sof\sdeleted\sfiles:\s(?P<files_deleted>[,0123456789]+)$')
    stat_ex_files_transferred_re = re.compile(r'Number\sof\sregular\sfiles\stransferred:\s(?P<files_transferred>[,0123456789]+)$')
    stat_ex_files_transferred_size_re = re.compile(r'Total\sfile\ssize:\s(?P<files_transferred_size>[,.0123456789]+\S?)\sbytes$')
    stat_ex_literal_data_re = re.compile(r'Literal\sdata:\s(?P<literal_data>[,.0123456789]+\S?)\sbytes$')
    stat_ex_matched_data_re = re.compile(r'Matched\sdata:\s(?P<matched_data>[,.0123456789]+\S?)\sbytes$')
    stat_ex_file_list_size_re = re.compile(r'File\slist\ssize:\s(?P<file_list_size>[,.0123456789]+\S?)$')
    stat_ex_file_list_generation_time_re = re.compile(r'File\slist\sgeneration\stime:\s(?P<file_list_generation_time>[,.0123456789]+\S?)\sseconds$')
    stat_ex_file_list_transfer_time_re = re.compile(r'File\slist\stransfer\stime:\s(?P<file_list_transfer_time>[,.0123456789]+\S?)\sseconds$')

    if jc.utils.has_data(data):
        for line in filter(None, data.splitlines()):

            file_line = file_line_re.match(line)
            if file_line:
                filename = file_line.group('name')
                meta = file_line.group('meta')

                output_line = {
                    'filename': filename,
                    'metadata': meta,
                    'update_type': update_type[meta[0]],
                    'file_type': file_type[meta[1]],
                    'checksum_or_value_different': checksum_or_value_different[meta[2]],
                    'size_different': size_different[meta[3]],
                    'modification_time_different': modification_time_different[meta[4]],
                    'permissions_different': permissions_different[meta[5]],
                    'owner_different': owner_different[meta[6]],
                    'group_different': group_different[meta[7]],
                    'acl_different': acl_different[meta[9]],
                    'extended_attribute_different': extended_attribute_different[meta[10]]
                }

                rsync_run['files'].append(output_line)
                continue

            file_line_mac = file_line_mac_re.match(line)
            if file_line_mac:
                filename = file_line_mac.group('name')
                meta = file_line_mac.group('meta')

                output_line = {
                    'filename': filename,
                    'metadata': meta,
                    'update_type': update_type[meta[0]],
                    'file_type': file_type[meta[1]],
                    'checksum_or_value_different': checksum_or_value_different[meta[2]],
                    'size_different': size_different[meta[3]],
                    'modification_time_different': modification_time_different[meta[4]],
                    'permissions_different': permissions_different[meta[5]],
                    'owner_different': owner_different[meta[6]],
                    'group_different': group_different[meta[7]]
                }

                rsync_run['files'].append(output_line)
                continue

            file_line_log = file_line_log_re.match(line)
            if file_line_log:
                filename = file_line_log.group('name')
                date = file_line_log.group('date')
                time = file_line_log.group('time')
                process = file_line_log.group('process')
                meta = file_line_log.group('meta')

                if process != last_process:
                    raw_output.append(rsync_run)
                    rsync_run = deepcopy(rsync_run_new)
                    last_process = process

                output_line = {
                    'filename': filename,
                    'date': date,
                    'time': time,
                    'process': process,
                    'metadata': meta,
                    'update_type': update_type[meta[0]],
                    'file_type': file_type[meta[1]],
                    'checksum_or_value_different': checksum_or_value_different[meta[2]],
                    'size_different': size_different[meta[3]],
                    'modification_time_different': modification_time_different[meta[4]],
                    'permissions_different': permissions_different[meta[5]],
                    'owner_different': owner_different[meta[6]],
                    'group_different': group_different[meta[7]],
                    'acl_different': acl_different[meta[9]],
                    'extended_attribute_different': extended_attribute_different[meta[10]]
                }
                rsync_run['files'].append(output_line)
                continue

            file_line_log_mac = file_line_log_mac_re.match(line)
            if file_line_log_mac:
                filename = file_line_log_mac.group('name')
                date = file_line_log_mac.group('date')
                time = file_line_log_mac.group('time')
                process = file_line_log_mac.group('process')
                meta = file_line_log_mac.group('meta')

                if process != last_process:
                    raw_output.append(rsync_run)
                    rsync_run = deepcopy(rsync_run_new)
                    last_process = process

                output_line = {
                    'filename': filename,
                    'date': date,
                    'time': time,
                    'process': process,
                    'metadata': meta,
                    'update_type': update_type[meta[0]],
                    'file_type': file_type[meta[1]],
                    'checksum_or_value_different': checksum_or_value_different[meta[2]],
                    'size_different': size_different[meta[3]],
                    'modification_time_different': modification_time_different[meta[4]],
                    'permissions_different': permissions_different[meta[5]],
                    'owner_different': owner_different[meta[6]],
                    'group_different': group_different[meta[7]]
                }
                rsync_run['files'].append(output_line)
                continue

            stat1_line = stat1_line_re.match(line)
            if stat1_line:
                rsync_run['summary'].update({
                    'sent': stat1_line.group('sent'),
                    'received': stat1_line.group('received'),
                    'bytes_sec': stat1_line.group('bytes_sec')
                })
                continue

            stat2_line = stat2_line_re.match(line)
            if stat2_line:
                rsync_run['summary']['total_size'] = stat2_line.group('total_size')
                rsync_run['summary']['speedup'] = stat2_line.group('speedup')
                continue

            stat1_line_simple = stat1_line_simple_re.match(line)
            if stat1_line_simple:
                rsync_run['summary'].update({
                    'sent': stat1_line_simple.group('sent'),
                    'received': stat1_line_simple.group('received'),
                    'bytes_sec': stat1_line_simple.group('bytes_sec')
                })
                continue

            stat2_line_simple = stat2_line_simple_re.match(line)
            if stat2_line_simple:
                rsync_run['summary']['total_size'] = stat2_line_simple.group('total_size')
                rsync_run['summary']['speedup'] = stat2_line_simple.group('speedup')
                continue

            stat_line_log = stat_line_log_re.match(line)
            if stat_line_log:
                rsync_run['summary'].update({
                    'date': stat_line_log.group('date'),
                    'time': stat_line_log.group('time'),
                    'process': stat_line_log.group('process'),
                    'sent': stat_line_log.group('sent'),
                    'received': stat_line_log.group('received'),
                    'total_size': stat_line_log.group('total_size')
                })
                continue

            stat1_line_log_v = stat1_line_log_v_re.match(line)
            if stat1_line_log_v:
                rsync_run['summary'].update({
                    'date': stat1_line_log_v.group('date'),
                    'time': stat1_line_log_v.group('time'),
                    'process': stat1_line_log_v.group('process'),
                    'matches': stat1_line_log_v.group('matches'),
                    'hash_hits': stat1_line_log_v.group('hash_hits'),
                    'false_alarms': stat1_line_log_v.group('false_alarms'),
                    'data': stat1_line_log_v.group('data')
                })
                continue

            stat2_line_log_v = stat2_line_log_v_re.match(line)
            if stat2_line_log_v:
                rsync_run['summary']['sent'] = stat2_line_log_v.group('sent')
                rsync_run['summary']['received'] = stat2_line_log_v.group('received')
                rsync_run['summary']['bytes_sec'] = stat2_line_log_v.group('bytes_sec')
                continue

            stat3_line_log_v = stat3_line_log_v_re.match(line)
            if stat3_line_log_v:
                rsync_run['summary']['total_size'] = stat3_line_log_v.group('total_size')
                rsync_run['summary']['speedup'] = stat3_line_log_v.group('speedup')
                continue

            # extra stats lines when using rsync --stats or --info=stats[1-3]
            stat_ex_files_number_v = stat_ex_files_number_re.match(line)
            if stat_ex_files_number_v:
                rsync_run['summary']['total_files'] = stat_ex_files_number_v.group('files_total')
                rsync_run['summary']['regular_files'] = stat_ex_files_number_v.group('files_regular')
                rsync_run['summary']['dir_files'] = stat_ex_files_number_v.group('files_dir')
                continue

            stat_ex_files_created_v = stat_ex_files_created_re.match(line)
            if stat_ex_files_created_v:
                rsync_run['summary']['total_created_files'] = stat_ex_files_created_v.group('files_created_total')
                rsync_run['summary']['created_regular_files'] = stat_ex_files_created_v.group('files_created_regular')
                rsync_run['summary']['created_dir_files'] = stat_ex_files_created_v.group('files_created_dir')
                continue

            stat_ex_files_deleted_v = stat_ex_files_deleted_re.match(line)
            if stat_ex_files_deleted_v:
                rsync_run['summary']['deleted_files'] = stat_ex_files_deleted_v.group('files_deleted')
                continue

            stat_ex_files_transferred_v = stat_ex_files_transferred_re.match(line)
            if stat_ex_files_transferred_v:
                rsync_run['summary']['transferred_files'] = stat_ex_files_transferred_v.group('files_transferred')
                continue

            stat_ex_files_transferred_size_v = stat_ex_files_transferred_size_re.match(line)
            if stat_ex_files_transferred_size_v:
                rsync_run['summary']['transferred_file_size'] = stat_ex_files_transferred_size_v.group('files_transferred_size')
                continue

            stat_ex_literal_data_v = stat_ex_literal_data_re.match(line)
            if stat_ex_literal_data_v:
                rsync_run['summary']['literal_data'] = stat_ex_literal_data_v.group('literal_data')
                continue

            stat_ex_matched_data_v = stat_ex_matched_data_re.match(line)
            if stat_ex_matched_data_v:
                rsync_run['summary']['matched_data'] = stat_ex_matched_data_v.group('matched_data')
                continue

            stat_ex_file_list_size_v = stat_ex_file_list_size_re.match(line)
            if stat_ex_file_list_size_v:
                rsync_run['summary']['file_list_size'] = stat_ex_file_list_size_v.group('file_list_size')
                continue

            stat_ex_file_list_generation_time_v = stat_ex_file_list_generation_time_re.match(line)
            if stat_ex_file_list_generation_time_v:
                rsync_run['summary']['file_list_generation_time'] = stat_ex_file_list_generation_time_v.group('file_list_generation_time')
                continue

            stat_ex_file_list_transfer_time_v = stat_ex_file_list_transfer_time_re.match(line)
            if stat_ex_file_list_transfer_time_v:
                rsync_run['summary']['file_list_transfer_time'] = stat_ex_file_list_transfer_time_v.group('file_list_transfer_time')
                continue

    raw_output.append(rsync_run)

    # cleanup blank entries
    raw_output = [run for run in raw_output if run != rsync_run_new]

    return raw_output if raw else _process(raw_output)



r"""jc - JSON Convert M3U and M3U8 file parser

This parser will make a best-effort to parse extended field information. If
the extended fields cannot be successfully parsed, then an `unparsed_info`
field will be added to the object. If not using `--quiet`, then a warning
message also will be printed to `STDERR`.

Usage (cli):

    $ cat playlist.m3u | jc --m3u

Usage (module):

    import jc
    result = jc.parse('m3u', m3u_file_output)

Schema:

    [
      {
        "runtime":              integer,
        "display":              string,
        "path":                 string,
        <extended fields>:      string,  # [0]
        "unparsed_info":        string,  # [1]
      }
    ]

    [0] Field names are pulled directly from the #EXTINF: line
    [1] Only added if the extended information cannot be parsed

Examples:

    $ cat playlist.m3u | jc --m3u -p
    [
      {
        "runtime": 105,
        "display": "Example artist - Example title",
        "path": "C:\\Files\\My Music\\Example.mp3"
      },
      {
        "runtime": 321,
        "display": "Example Artist2 - Example title2",
        "path": "C:\\Files\\My Music\\Favorites\\Example2.ogg"
      }
    ]

    $ cat playlist.m3u | jc --m3u -p -r
    [
      {
        "runtime": "105",
        "display": "Example artist - Example title",
        "path": "C:\\Files\\My Music\\Example.mp3"
      },
      {
        "runtime": "321",
        "display": "Example Artist2 - Example title2",
        "path": "C:\\Files\\My Music\\Favorites\\Example2.ogg"
      }
    ]
"""
import shlex
from typing import List, Dict
import jc.utils


class info():
    """Provides parser metadata (version, author, etc.)"""
    version = '1.0'
    description = 'M3U and M3U8 file parser'
    author = 'Kelly Brazil'
    author_email = 'kellyjonbrazil@gmail.com'
    compatible = ['linux', 'darwin', 'cygwin', 'win32', 'aix', 'freebsd']
    tags = ['file']


__version__ = info.version


def _process(proc_data: List[Dict]) -> List[Dict]:
    """
    Final processing to conform to the schema.

    Parameters:

        proc_data:   (List of Dictionaries) raw structured data to process

    Returns:

        List of Dictionaries. Structured to conform to the schema.
    """
    int_list = {'runtime'}

    for entry in proc_data:
        for key in entry:
            if key in int_list:
                entry[key] = jc.utils.convert_to_int(entry[key])

    return proc_data


def parse(
    data: str,
    raw: bool = False,
    quiet: bool = False
) -> List[Dict]:
    """
    Main text parsing function

    Parameters:

        data:        (string)  text data to parse
        raw:         (boolean) unprocessed output if True
        quiet:       (boolean) suppress warning messages if True

    Returns:

        List of Dictionaries. Raw or processed structured data.
    """
    jc.utils.compatibility(__name__, info.compatible, quiet)
    jc.utils.input_type_check(data)

    raw_output: List = []
    output_line = {}

    if jc.utils.has_data(data):
        for line in filter(None, data.splitlines()):
            # ignore any lines with only whitespace
            if not jc.utils.has_data(line):
                continue

            # extended info fields
            if line.lstrip().startswith('#EXTINF:'):
                splitline = line.strip().split(':', maxsplit=1)

                # best-effort to parse additional extended fields
                # if a parsing error occurs, a warning message will be
                # printed to STDERR and `unparsed_info` added
                try:
                    extline = shlex.shlex(splitline[1], posix=True)
                    extline.whitespace_split = True
                    extline.whitespace = ', '  # add comma to whitespace detection
                    extline.quotes = '"'  # remove single quotes
                    extline_list = list(extline)
                    runtime = extline_list.pop(0)
                    display_list = []

                    for item in extline_list:
                        if '=' in item:
                            k, v = item.split('=', maxsplit=1)
                            output_line.update({k: v})

                        else:
                            display_list.append(item)

                    display = ' '.join(display_list)
                    output_line.update({
                        'runtime': runtime,
                        'display': display
                    })

                except Exception:
                    if not quiet:
                        jc.utils.warning_message([
                            'Not able to parse non-standard extensions in the following line:',
                            line
                        ])
                    output_line = {'unparsed_info': line}

                continue

            # ignore all other extension info (obsolete)
            if line.lstrip().startswith('#'):
                continue

            # any lines left over are paths
            output_line.update(
                {'path': line.strip()}
            )

            raw_output.append(output_line)
            output_line = {}

    return raw_output if raw else _process(raw_output)



import unittest
import os
import sys

sys.path.append(os.getcwd())
from tests import utils_for_test as test_utils

sys.path.pop()


class MyTests(unittest.TestCase):
    def test_wg_show_nodata(self):
        """
        Test 'wg-show' with no data
        """
        test_utils.run_no_data(self, __file__, [])

    def test_wg_show_all_fixtures(self):
        """
        Test 'wg-show' with various fixtures
        """
        test_utils.run_all_fixtures(self, __file__)


if __name__ == "__main__":
    unittest.main()



r"""jc - JSON Convert `tsv` file parser

The `tsv` parser is a clone of the `csv` parser that uses '\t'
as the delimiter character. The first row of the file must be a header row.

Usage (cli):

    $ cat file.tsv | jc --tsv

Usage (module):

    import jc
    result = jc.parse('tsv', tsv_output)

Schema:

TSV file converted to a Dictionary:
https://docs.python.org/3/library/csv.html

    [
      {
        "column_name1":     string,
        "column_name2":     string
      }
    ]

Examples:

    $ cat homes.tsv
    "Sell"	"List"	"Living"	"Rooms"	"Beds"	"Baths"	"Age"	"Acres"...
    142	160	28	10	5	3	60	0.28	3167
    175	180	18	8	4	1	12	0.43	4033
    129	132	13	6	3	1	41	0.33	1471
    ...

    $ cat homes.tsv | jc --tsv -p
    [
      {
        "Sell": "142",
        "List": "160",
        "Living": "28",
        "Rooms": "10",
        "Beds": "5",
        "Baths": "3",
        "Age": "60",
        "Acres": "0.28",
        "Taxes": "3167"
      },
      {
        "Sell": "175",
        "List": "180",
        "Living": "18",
        "Rooms": "8",
        "Beds": "4",
        "Baths": "1",
        "Age": "12",
        "Acres": "0.43",
        "Taxes": "4033"
      },
      {
        "Sell": "129",
        "List": "132",
        "Living": "13",
        "Rooms": "6",
        "Beds": "3",
        "Baths": "1",
        "Age": "41",
        "Acres": "0.33",
        "Taxes": "1471"
      },
      ...
    ]
"""
from typing import List, Union
from jc.jc_types import JSONDictType
import jc.parsers.csv as jc_csv
import jc.utils


class info():
    """Provides parser metadata (version, author, etc.)"""
    version = '1.0'
    description = 'TSV file parser'
    author = 'Gary Gurlaskie'
    author_email = 'https://github.com/garyg1'
    details = 'Using the python standard csv library'
    compatible = ['linux', 'darwin', 'cygwin', 'win32', 'aix', 'freebsd']
    tags = ['standard', 'file', 'string']


__version__ = info.version


def parse(
    data: Union[str, bytes],
    raw: bool = False,
    quiet: bool = False
) -> List[JSONDictType]:
    """
    Main text parsing function

    Parameters:

        data:        (string)  text data to parse
        raw:         (boolean) unprocessed output if True
        quiet:       (boolean) suppress warning messages if True

    Returns:

        List of Dictionaries. Raw or processed structured data.
    """
    jc.utils.compatibility(__name__, info.compatible, quiet)
    jc.utils.input_type_check(data)
    return jc_csv.parse(data, raw, quiet, implicit_header=False, tsv=True)


import os
import unittest
import json
import jc.parsers.cef

THIS_DIR = os.path.dirname(os.path.abspath(__file__))


class MyTests(unittest.TestCase):

    # input
    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/cef.out'), 'r', encoding='utf-8') as f:
        cef = f.read()

    # output
    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/cef.json'), 'r', encoding='utf-8') as f:
        cef_json = json.loads(f.read())


    def test_cef_nodata(self):
        """
        Test 'cef' with no data
        """
        self.assertEqual(jc.parsers.cef.parse('', quiet=True), [])

    def test_cef_sample(self):
        """
        Test with sample cef log
        """
        self.assertEqual(jc.parsers.cef.parse(self.cef, quiet=True), self.cef_json)


if __name__ == '__main__':
    unittest.main()



"""
Module for working with PNP ID REGISTRY
"""

import csv
import string
from html.parser import HTMLParser
from urllib import request

__all__ = ["Registry"]


class WebPnpIdParser(HTMLParser):
    """Parser pnp id from https://uefi.org/PNP_ID_List

    Examples:
        p = WebPnpIdParser()
        p.feed(html_data)
        p.result
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._find_table = False
        self._find_row = False
        # first -- company name, second -- pnp id, third -- approved date
        self._last_field = []
        # key -- pnp id, value -- tuple (company_name, approved_date)
        self.result = {}

    def handle_starttag(self, tag, attrs):
        if tag == "tbody":
            self._find_table = True
        elif self._find_table and tag == "tr":
            self._find_row = True

    def handle_endtag(self, tag):
        if tag == "tbody":
            self._find_table = False
        elif self._find_table and tag == "tr":
            self._find_row = False
            # add table row to result
            self.result[self._last_field[1]] = (
                self._last_field[0],
                self._last_field[-1],
            )
            self._last_field.clear()

    def handle_data(self, data):
        # skip processing until table is found
        if not self._find_table:
            return

        if self._find_row:
            data = data.strip()
            if data:
                self._last_field.append(data)

    def error(self, message):
        super().close()


class Registry(dict):
    """Registry pnp id data dictionary

    key   -- pnp_id
    value -- company name
    """

    @classmethod
    def from_web(cls, filter_by_id: str = None):
        """Get registry from https://uefi.org/PNP_ID_List

        Args:
            filter_by_id (str), optional: filter registry by id

        Raises:

        Returns:

        """
        url = "https://uefi.org/PNP_ID_List"
        if filter_by_id:
            url += "?search={}".format(filter_by_id)

        with request.urlopen(url) as req:
            parse = WebPnpIdParser()
            parse.feed(req.read().decode())

            registry = cls()
            for key, value in parse.result.items():
                # skip invalid search value
                if filter_by_id and key != filter_by_id:
                    continue
                registry[key] = value[0]
        return registry

    @classmethod
    def from_csv(cls, csv_path: str, filter_by_id: str = None):
        """Get registry by csv local file

        Args:
            csv_path (str): path to csv file
            filter_by_id (str), optional: filter registry by id

        Raises:

        Returns:

        """
        registry = cls()
        with open(csv_path, "r") as file:
            reader = csv.reader(file)
            for line in reader:
                # filter
                if filter_by_id and filter_by_id != line[0]:
                    continue
                registry[line[0]] = line[1]
        return registry

    def to_csv(self, csv_path: str):
        """Dump registry to csv file"""
        with open(csv_path, "w") as csv_file:
            writer = csv.writer(csv_file)
            writer.writerows(self.items())
        return self

    def get_company_from_id(self, pnp_id: str) -> str:
        """Convert PNP id to company name"""
        return self.get(pnp_id, "Unknown")

    def get_company_from_raw(self, raw: int) -> str:
        """Convert raw edid value to company name"""
        tmp = [(raw >> 10) & 31, (raw >> 5) & 31, raw & 31]
        pnp_id = "".join(string.ascii_uppercase[n - 1] for n in tmp)
        return self.get_company_from_id(pnp_id)



import os
import unittest
import json
import jc.parsers.rsync

THIS_DIR = os.path.dirname(os.path.abspath(__file__))


class MyTests(unittest.TestCase):

    # input
    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/rsync-i.out'), 'r', encoding='utf-8') as f:
        centos_7_7_rsync_i = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/rsync-i.out'), 'r', encoding='utf-8') as f:
        generic_rsync_i = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/rsync-ivvv.out'), 'r', encoding='utf-8') as f:
        centos_7_7_rsync_ivvv = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/osx-10.14.6/rsync-i-vvv.out'), 'r', encoding='utf-8') as f:
        osx_10_14_6_rsync_ivvv = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/rsync-ivvv-nochange.out'), 'r', encoding='utf-8') as f:
        centos_7_7_rsync_ivvv_nochange = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/osx-10.14.6/rsync-i-vvv-nochange.out'), 'r', encoding='utf-8') as f:
        osx_10_14_6_rsync_ivvv_nochange = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/rsync-i-logfile.out'), 'r', encoding='utf-8') as f:
        centos_7_7_rsync_i_logfile = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/rsync-i-v-logfile.out'), 'r', encoding='utf-8') as f:
        centos_7_7_rsync_i_v_logfile = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/rsync-i-vv-logfile.out'), 'r', encoding='utf-8') as f:
        centos_7_7_rsync_i_vv_logfile = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/rsync-i-vvv-logfile.out'), 'r', encoding='utf-8') as f:
        centos_7_7_rsync_i_vvv_logfile = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/rsync-i-vvv-logfile-nochange.out'), 'r', encoding='utf-8') as f:
        centos_7_7_rsync_i_vvv_logfile_nochange = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/osx-10.14.6/rsync-i-vvv-logfile-nochange.out'), 'r', encoding='utf-8') as f:
        osx_10_14_6_rsync_i_vvv_logfile_nochange = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/rsync-i-stats.out'), 'r', encoding='utf-8') as f:
        generic_rsync_i_stats = f.read()

    # output
    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/rsync-i.json'), 'r', encoding='utf-8') as f:
        centos_7_7_rsync_i_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/rsync-i.json'), 'r', encoding='utf-8') as f:
        generic_rsync_i_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/rsync-ivvv.json'), 'r', encoding='utf-8') as f:
        centos_7_7_rsync_ivvv_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/osx-10.14.6/rsync-i-vvv.json'), 'r', encoding='utf-8') as f:
        osx_10_14_6_rsync_ivvv_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/rsync-ivvv-nochange.json'), 'r', encoding='utf-8') as f:
        centos_7_7_rsync_ivvv_nochange_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/osx-10.14.6/rsync-i-vvv-nochange.json'), 'r', encoding='utf-8') as f:
        osx_10_14_6_rsync_ivvv_nochange_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/rsync-i-logfile.json'), 'r', encoding='utf-8') as f:
        centos_7_7_rsync_i_logfile_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/rsync-i-v-logfile.json'), 'r', encoding='utf-8') as f:
        centos_7_7_rsync_i_v_logfile_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/rsync-i-vv-logfile.json'), 'r', encoding='utf-8') as f:
        centos_7_7_rsync_i_vv_logfile_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/rsync-i-vvv-logfile.json'), 'r', encoding='utf-8') as f:
        centos_7_7_rsync_i_vvv_logfile_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/centos-7.7/rsync-i-vvv-logfile-nochange.json'), 'r', encoding='utf-8') as f:
        centos_7_7_rsync_i_vvv_logfile_nochange_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/osx-10.14.6/rsync-i-vvv-logfile-nochange.json'), 'r', encoding='utf-8') as f:
        osx_10_14_6_rsync_i_vvv_logfile_nochange_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/rsync-i-stats.json'), 'r', encoding='utf-8') as f:
        generic_rsync_i_stats_json = json.loads(f.read())


    def test_rsync_nodata(self):
        """
        Test 'rsync' with no data
        """
        self.assertEqual(jc.parsers.rsync.parse('', quiet=True), [])

    def test_rsync_centos_7_7_rsync_i(self):
        """
        Test 'rsync -i' on Centos 7.7
        """
        self.assertEqual(jc.parsers.rsync.parse(self.centos_7_7_rsync_i, quiet=True), self.centos_7_7_rsync_i_json)

    def test_rsync_generic_rsync_i(self):
        """
        Test 'rsync -i'
        """
        self.assertEqual(jc.parsers.rsync.parse(self.generic_rsync_i, quiet=True), self.generic_rsync_i_json)

    def test_rsync_centos_7_7_rsync_ivvv(self):
        """
        Test 'rsync -ivvv' on Centos 7.7
        """
        self.assertEqual(jc.parsers.rsync.parse(self.centos_7_7_rsync_ivvv, quiet=True), self.centos_7_7_rsync_ivvv_json)

    def test_rsync_osx_10_14_6_rsync_ivvv(self):
        """
        Test 'rsync -ivvv' on OSX 10.14.6
        """
        self.assertEqual(jc.parsers.rsync.parse(self.osx_10_14_6_rsync_ivvv, quiet=True), self.osx_10_14_6_rsync_ivvv_json)

    def test_rsync_osx_10_14_6_rsync_ivvv_nochange(self):
        """
        Test 'rsync -ivvv' on OSX 10.14.6 with no file changes
        """
        self.assertEqual(jc.parsers.rsync.parse(self.osx_10_14_6_rsync_ivvv_nochange, quiet=True), self.osx_10_14_6_rsync_ivvv_nochange_json)

    def test_rsync_centos_7_7_rsync_ivvv_nochange(self):
        """
        Test 'rsync -ivvv' on Centos 7.7 with no file changes
        """
        self.assertEqual(jc.parsers.rsync.parse(self.centos_7_7_rsync_ivvv_nochange, quiet=True), self.centos_7_7_rsync_ivvv_nochange_json)

    def test_rsync_centos_7_7_rsync_i_logfile(self):
        """
        Test 'rsync -i --log-file=xxx' on Centos 7.7
        """
        self.assertEqual(jc.parsers.rsync.parse(self.centos_7_7_rsync_i_logfile, quiet=True), self.centos_7_7_rsync_i_logfile_json)

    def test_rsync_centos_7_7_rsync_i_v_logfile(self):
        """
        Test 'rsync -iv --log-file=xxx' on Centos 7.7
        """
        self.assertEqual(jc.parsers.rsync.parse(self.centos_7_7_rsync_i_v_logfile, quiet=True), self.centos_7_7_rsync_i_v_logfile_json)

    def test_rsync_centos_7_7_rsync_i_vv_logfile(self):
        """
        Test 'rsync -ivv --log-file=xxx' on Centos 7.7
        """
        self.assertEqual(jc.parsers.rsync.parse(self.centos_7_7_rsync_i_vv_logfile, quiet=True), self.centos_7_7_rsync_i_vv_logfile_json)

    def test_rsync_centos_7_7_rsync_i_vvv_logfile(self):
        """
        Test 'rsync -ivvv --log-file=xxx' on Centos 7.7
        """
        self.assertEqual(jc.parsers.rsync.parse(self.centos_7_7_rsync_i_vvv_logfile, quiet=True), self.centos_7_7_rsync_i_vvv_logfile_json)

    def test_rsync_centos_7_7_rsync_i_vvv_logfile_nochange(self):
        """
        Test 'rsync -ivvv --log-file=xxx' on Centos 7.7 with no file changes
        """
        self.assertEqual(jc.parsers.rsync.parse(self.centos_7_7_rsync_i_vvv_logfile_nochange, quiet=True), self.centos_7_7_rsync_i_vvv_logfile_nochange_json)

    def test_rsync_osx_10_14_6_rsync_i_vvv_logfile_nochange(self):
        """
        Test 'rsync -ivvv --log-file=xxx' on OSX 10.14.6 with no file changes
        """
        self.assertEqual(jc.parsers.rsync.parse(self.osx_10_14_6_rsync_i_vvv_logfile_nochange, quiet=True), self.osx_10_14_6_rsync_i_vvv_logfile_nochange_json)

    def test_rsync_simple_summary(self):
        """
        Test 'rsync avh' output with a simple summary
        """
        data = '''sending incremental file list

sent 8.71M bytes  received 29.88K bytes  10.99K bytes/sec
total size is 221.79G  speedup is 25,388.23
'''
        expected = [{"summary":{"sent":8710000,"received":29880,"bytes_sec":10990.0,"total_size":221790000000,"speedup":25388.23},"files":[]}]
        self.assertEqual(jc.parsers.rsync.parse(data, quiet=True), expected)

    def test_rsync_with_stats(self):
        """
        Test 'rsync -i --stats' or 'rsync -i --info=stats[1-3]'
        """
        self.assertEqual(jc.parsers.rsync.parse(self.generic_rsync_i_stats, quiet=True), self.generic_rsync_i_stats_json)


if __name__ == '__main__':
    unittest.main()



r"""jc - JSON Convert `/proc/pagetypeinfo` file parser

Usage (cli):

    $ cat /proc/pagetypeinfo | jc --proc

or

    $ jc /proc/pagetypeinfo

or

    $ cat /proc/pagetypeinfo | jc --proc-pagetypeinfo

Usage (module):

    import jc
    result = jc.parse('proc', proc_pagetypeinfo_file)

or

    import jc
    result = jc.parse('proc_pagetypeinfo', proc_pagetypeinfo_file)

Schema:

    {
      "page_block_order":             integer,
      "pages_per_block":              integer,
      "free_pages": [
        {
          "node":                     integer,
          "zone":                     string,
          "type":                     string,
          "free": [
                                      integer  # [0]
          ]
      ],
      "num_blocks_type": [
        {
          "node":                     integer,
          "zone":                     string,
          "unmovable":                integer,
          "movable":                  integer,
          "reclaimable":              integer,
          "high_atomic":              integer,
          "isolate":                  integer
        }
      ]
    }

    [0] array index correlates to the Order number.
        E.g. free[0] is the value for Order 0

Examples:

    $ cat /proc/pagetypeinfo | jc --proc -p
    {
      "page_block_order": 9,
      "pages_per_block": 512,
      "free_pages": [
        {
          "node": 0,
          "zone": "DMA",
          "type": "Unmovable",
          "free": [
            0,
            0,
            0,
            1,
            1,
            1,
            1,
            1,
            0,
            0,
            0
          ]
        },
        ...
      ],
      "num_blocks_type": [
        {
          "node": 0,
          "zone": "DMA",
          "unmovable": 1,
          "movable": 7,
          "reclaimable": 0,
          "high_atomic": 0,
          "isolate": 0
        },
        {
          "node": 0,
          "zone": "DMA32",
          "unmovable": 8,
          "movable": 1472,
          "reclaimable": 48,
          "high_atomic": 0,
          "isolate": 0
        },
        {
          "node": 0,
          "zone": "Normal",
          "unmovable": 120,
          "movable": 345,
          "reclaimable": 47,
          "high_atomic": 0,
          "isolate": 0
        }
      ]
    }
"""
from typing import Dict
import jc.utils


class info():
    """Provides parser metadata (version, author, etc.)"""
    version = '1.0'
    description = '`/proc/pagetypeinfo` file parser'
    author = 'Kelly Brazil'
    author_email = 'kellyjonbrazil@gmail.com'
    compatible = ['linux']
    tags = ['file']
    hidden = True


__version__ = info.version


def _process(proc_data: Dict) -> Dict:
    """
    Final processing to conform to the schema.

    Parameters:

        proc_data:   (Dictionary) raw structured data to process

    Returns:

        Dictionary. Structured to conform to the schema.
    """
    return proc_data


def parse(
    data: str,
    raw: bool = False,
    quiet: bool = False
) -> Dict:
    """
    Main text parsing function

    Parameters:

        data:        (string)  text data to parse
        raw:         (boolean) unprocessed output if True
        quiet:       (boolean) suppress warning messages if True

    Returns:

        Dictionary. Raw or processed structured data.
    """
    jc.utils.compatibility(__name__, info.compatible, quiet)
    jc.utils.input_type_check(data)

    raw_output: Dict = {}
    section = ''

    if jc.utils.has_data(data):

        for line in filter(None, data.splitlines()):

            if line.startswith('Page block order:'):
                raw_output['page_block_order'] = int(line.split(':', maxsplit=1)[1])
                continue

            if line.startswith('Pages per block:'):
                raw_output['pages_per_block'] = int(line.split(':', maxsplit=1)[1])
                continue

            if line.startswith('Free pages count per migrate type at order'):
                section = 'free_pages'
                raw_output['free_pages'] = []
                continue

            if line.startswith('Number of blocks type'):
                section = 'num_blocks_type'
                raw_output['num_blocks_type'] = []
                continue

            # Free pages count per migrate type at order       0      1      2      3      4      5      6      7      8      9     10
            # Node    0, zone      DMA, type    Unmovable      0      0      0      1      1      1      1      1      0      0      0
            if section == 'free_pages':
                split_line = line.replace(',', ' ').split()

                output_line = {
                    'node': int(split_line[1]),
                    'zone': split_line[3],
                    'type': split_line[5],
                    'free': [int(x) for x in split_line[6:]]
                }

                raw_output['free_pages'].append(output_line)
                continue

            # Number of blocks type     Unmovable      Movable  Reclaimable   HighAtomic      Isolate 
            # Node 0, zone      DMA            1            7            0            0            0 
            if section == 'num_blocks_type':
                split_line = line.replace(',', ' ').split()

                output_line = {
                    'node': int(split_line[1]),
                    'zone': split_line[3],
                    'unmovable': int(split_line[4]),
                    'movable': int(split_line[5]),
                    'reclaimable': int(split_line[6]),
                    'high_atomic': int(split_line[7]),
                    'isolate': int(split_line[8]),
                }

                raw_output['num_blocks_type'].append(output_line)
                continue

    return raw_output if raw else _process(raw_output)



r"""jc - JSON Convert `/proc/<pid>/fdinfo/<fd>` file parser

Usage (cli):

    $ cat /proc/1/fdinfo/5 | jc --proc

or

    $ jc /proc/1/fdinfo/5

or

    $ cat /proc/1/fdinfo/5 | jc --proc-pid-fdinfo

Usage (module):

    import jc
    result = jc.parse('proc', proc_pid_fdinfo_file)

or

    import jc
    result = jc.parse('proc_pid_fdinfo', proc_pid_fdinfo_file)

Schema:

Any unspecified fields are strings.

    {
      "pos":                        integer,
      "flags":                      integer,
      "mnt_id":                     integer,
      "scm_fds":                    string,
      "ino":                        integer,
      "lock":                       string,
      "epoll": {
        "tfd":                      integer,
        "events":                   string,
        "data":                     string,
        "pos":                      integer,
        "ino":                      string,
        "sdev":                     string
      },
      "inotify": {
        "wd":                       integer,
        "ino":                      string,
        "sdev":                     string,
        "mask":                     string,
        "ignored_mask":             string,
        "fhandle-bytes":            string,
        "fhandle-type":             string,
        "f_handle":                 string
      },
      "fanotify": {
        "flags":                    string,
        "event-flags":              string,
        "mnt_id":                   string,
        "mflags":                   string,
        "mask":                     string,
        "ignored_mask":             string,
        "ino":                      string,
        "sdev":                     string,
        "fhandle-bytes":            string,
        "fhandle-type":             string,
        "f_handle":                 string
      },
      "clockid":                    integer,
      "ticks":                      integer,
      "settime flags":              integer,
      "it_value": [
                                    integer
      ],
      "it_interval": [
                                    integer
      ]
    }

Examples:

    $ cat /proc/1/fdinfo/5 | jc --proc -p
    {
      "pos": 0,
      "flags": 2,
      "mnt_id": 9,
      "ino": 63107,
      "clockid": 0,
      "ticks": 0,
      "settime flags": 1,
      "it_value": [
        0,
        49406829
      ],
      "it_interval": [
        1,
        0
      ]
    }
"""
import re
from typing import Dict
import jc.utils


class info():
    """Provides parser metadata (version, author, etc.)"""
    version = '1.0'
    description = '`/proc/<pid>/fdinfo/<fd>` file parser'
    author = 'Kelly Brazil'
    author_email = 'kellyjonbrazil@gmail.com'
    compatible = ['linux']
    tags = ['file']
    hidden = True


__version__ = info.version


def _process(proc_data: Dict) -> Dict:
    """
    Final processing to conform to the schema.

    Parameters:

        proc_data:   (List of Dictionaries) raw structured data to process

    Returns:

        Dictionary. Structured to conform to the schema.
    """
    root_int_list = {'pos', 'flags', 'mnt_id', 'ino', 'clockid', 'ticks',
                     'settime flags', 'size', 'count'}
    epoll_int_list = {'tfd', 'pos'}
    inotify_int_list = {'wd'}

    for key, val in proc_data.items():
        if key in root_int_list:
            proc_data[key] = int(val)

    if 'epoll' in proc_data:
        for key, val in proc_data['epoll'].items():
            if key in epoll_int_list:
                proc_data['epoll'][key] = int(val)

    if 'inotify' in proc_data:
        for key, val in proc_data['inotify'].items():
            if key in inotify_int_list:
                proc_data['inotify'][key] = int(val)

    return proc_data


def parse(
    data: str,
    raw: bool = False,
    quiet: bool = False
) -> Dict:
    """
    Main text parsing function

    Parameters:

        data:        (string)  text data to parse
        raw:         (boolean) unprocessed output if True
        quiet:       (boolean) suppress warning messages if True

    Returns:

        Dictionary. Raw or processed structured data.
    """
    jc.utils.compatibility(__name__, info.compatible, quiet)
    jc.utils.input_type_check(data)

    raw_output: Dict = {}
    split_me = {'it_value:', 'it_interval:'}

    if jc.utils.has_data(data):

        for line in filter(None, data.splitlines()):

            # epoll files
            if line.startswith('tfd:'):
                line_match = re.findall(r'(?P<key>\S+):(?:\s+)?(?P<val>\S+s*)', line)
                if line_match:
                    raw_output.update({'epoll': {k.strip(): v.strip() for k, v in line_match}})
                continue

            # inotify files
            if line.startswith('inotify'):
                split_line = line[8:].split()
                raw_output['inotify'] = {}
                for item in split_line:
                    k, v = item.split(':', maxsplit=1)
                    raw_output['inotify'][k] = v
                continue

            # fanotify files
            if line.startswith('fanotify'):
                split_line = line[9:].split()

                if not 'fanotify' in raw_output:
                    raw_output['fanotify'] = {}

                for item in split_line:
                    k, v = item.split(':', maxsplit=1)
                    raw_output['fanotify'][k] = v
                continue

            # timerfd files
            if line.split()[0] in split_me:
                split_line = line.replace(':', '').replace('(', '').replace(')', '').replace(',', '').split()
                raw_output[split_line[0]] = [int(x) for x in split_line[1:]]
                continue

            key, val = line.split(':', maxsplit=1)
            raw_output[key.strip()] = val.strip()
            continue

    return raw_output if raw else _process(raw_output)



r"""jc - JSON Convert `lspci -mmv` command output parser

This parser supports the following `lspci` options:
- `-mmv`
- `-nmmv`
- `-nnmmv`

Usage (cli):

    $ lspci -nnmmv | jc --lspci

or

    $ jc lspci -nnmmv

Usage (module):

    import jc
    result = jc.parse('lspci', lspci_command_output)

Schema:

    [
      {
        "slot":                         string,
        "domain":                       string,
        "domain_int":                   integer,
        "bus":                          string,
        "bus_int":                      integer,
        "dev":                          string,
        "dev_int":                      integer,
        "function":                     string,
        "function_int":                 integer,
        "class":                        string,
        "class_id":                     string,
        "class_id_int":                 integer,
        "vendor":                       string,
        "vendor_id":                    string,
        "vendor_id_int":                integer,
        "device":                       string,
        "device_id":                    string,
        "device_id_int":                integer,
        "svendor":                      string,
        "svendor_id":                   string,
        "svendor_id_int":               integer,
        "sdevice":                      string,
        "sdevice_id":                   string,
        "sdevice_id_int":               integer,
        "rev":                          string,
        "physlot":                      string,
        "progif":                       string,
        "progif_int":                   integer
      }
    ]

Examples:

    $ lspci -nnmmv | jc --lspci -p
    [
      {
        "slot": "ff:02:05.0",
        "domain": "ff",
        "domain_int": 255,
        "bus": "02",
        "bus_int": 2,
        "dev": "05",
        "dev_int": 5,
        "function": "0",
        "function_int": 0,
        "class": "SATA controller",
        "class_id": "0106",
        "class_id_int": 262,
        "vendor": "VMware",
        "vendor_id": "15ad",
        "vendor_id_int": 5549,
        "device": "SATA AHCI controller",
        "device_id": "07e0",
        "device_id_int": 2016,
        "svendor": "VMware",
        "svendor_id": "15ad",
        "svendor_id_int": 5549,
        "sdevice": "SATA AHCI controller",
        "sdevice_id": "07e0",
        "sdevice_id_int": 2016,
        "physlot": "37",
        "progif": "01",
        "progif_int": 1
      },
      ...
    ]

    $ lspci -nnmmv | jc --lspci -p -r
    [
      {
        "slot": "ff:02:05.0",
        "domain": "ff",
        "bus": "02",
        "dev": "05",
        "function": "0",
        "class": "SATA controller",
        "class_id": "0106",
        "vendor": "VMware",
        "vendor_id": "15ad",
        "device": "SATA AHCI controller",
        "device_id": "07e0",
        "svendor": "VMware",
        "svendor_id": "15ad",
        "sdevice": "SATA AHCI controller",
        "sdevice_id": "07e0",
        "physlot": "37",
        "progif": "01"
      },
      ...
    ]
"""
import re
from typing import List, Dict
from jc.jc_types import JSONDictType
import jc.utils


class info():
    """Provides parser metadata (version, author, etc.)"""
    version = '1.1'
    description = '`lspci -mmv` command parser'
    author = 'Kelly Brazil'
    author_email = 'kellyjonbrazil@gmail.com'
    compatible = ['linux']
    magic_commands = ['lspci']
    tags = ['command']


__version__ = info.version


def _process(proc_data: List[JSONDictType]) -> List[JSONDictType]:
    """
    Final processing to conform to the schema.

    Parameters:

        proc_data:   (List of Dictionaries) raw structured data to process

    Returns:

        List of Dictionaries. Structured to conform to the schema.
    """
    int_list: set[str] = {
        'domain', 'bus', 'dev', 'function', 'class_id', 'vendor_id', 'device_id',
        'svendor_id', 'sdevice_id', 'progif'
    }

    new_list: List[JSONDictType] = []

    for item in proc_data:
        output: Dict = {}
        for key, val in item.items():
            output[key] = val
            if key in int_list:
                output[key + '_int'] = int(val, 16)
        new_list.append(output)

    return new_list


def parse(
    data: str,
    raw: bool = False,
    quiet: bool = False
) -> List[JSONDictType]:
    """
    Main text parsing function

    Parameters:

        data:        (string)  text data to parse
        raw:         (boolean) unprocessed output if True
        quiet:       (boolean) suppress warning messages if True

    Returns:

        List of Dictionaries. Raw or processed structured data.
    """
    jc.utils.compatibility(__name__, info.compatible, quiet)
    jc.utils.input_type_check(data)

    raw_output: List = []
    device_output: Dict = {}

    if jc.utils.has_data(data):
        item_id_p = re.compile(r'(?P<id>^[0-9a-f]{4}$)')
        item_id_bracket_p = re.compile(r' \[(?P<id>[0-9a-f]{4})\]$')

        for line in filter(None, data.splitlines()):
            if line.startswith('Slot:'):
                if device_output:
                    raw_output.append(device_output)
                    device_output = {}

                device_output['slot'] = line.split()[1]

                slot_info = line.split()[1]
                *domain, bus, dev_fun = slot_info.split(':')

                if domain:
                    dom = domain[0]
                else:
                    dom = "00"

                dev, fun = dev_fun.split('.')
                device_output['domain'] = dom
                device_output['bus'] = bus
                device_output['dev'] = dev
                device_output['function'] = fun
                continue

            key, val = line.split(maxsplit=1)
            key = key[:-1].lower()

            # numeric only (-nmmv)
            if item_id_p.match(val):
                device_output[key + '_id'] = val
                continue

            # string and numeric (-nnmmv)
            if item_id_bracket_p.search(val):
                string, idnum = val.rsplit(maxsplit=1)
                device_output[key] = string
                device_output[key + '_id'] = idnum[1:-1]
                continue

            # string only (-mmv)
            device_output[key] = val
            continue


        if device_output:
            raw_output.append(device_output)

    return raw_output if raw else _process(raw_output)



import unittest
import json
import jc.parsers.url


class MyTests(unittest.TestCase):
    def test_url_nodata(self):
        """
        Test 'url' with no data
        """
        self.assertEqual(jc.parsers.url.parse('', quiet=True), {})


    def test_url_mailto(self):
        """
        Test mailto URL
        """
        data = r'<mailto:fred@example.com>'
        expected = json.loads(r'''{"url":"mailto:fred@example.com","scheme":"mailto","netloc":null,"path":"fred@example.com","parent":".","filename":"fred@example.com","stem":"fred@example","extension":"com","path_list":["fred@example.com"],"query":null,"query_obj":null,"fragment":null,"username":null,"password":null,"hostname":null,"port":null,"encoded":{"url":"mailto:fred@example.com","scheme":"mailto","netloc":null,"path":"fred@example.com","parent":".","filename":"fred@example.com","stem":"fred@example","extension":"com","path_list":["fred@example.com"],"query":null,"fragment":null,"username":null,"password":null,"hostname":null,"port":null},"decoded":{"url":"mailto:fred@example.com","scheme":"mailto","netloc":null,"path":"fred@example.com","parent":".","filename":"fred@example.com","stem":"fred@example","extension":"com","path_list":["fred@example.com"],"query":null,"fragment":null,"username":null,"password":null,"hostname":null,"port":null}}''')
        self.assertEqual(jc.parsers.url.parse(data, quiet=True), expected)


    def test_url_ftp(self):
        """
        Test ftp URL
        """
        data = r'ftp://localhost/filepath/filename.txt'
        expected = json.loads(r'''{"url":"ftp://localhost/filepath/filename.txt","scheme":"ftp","netloc":"localhost","path":"/filepath/filename.txt","parent":"/filepath","filename":"filename.txt","stem":"filename","extension":"txt","path_list":["filepath","filename.txt"],"query":null,"query_obj":null,"fragment":null,"username":null,"password":null,"hostname":"localhost","port":null,"encoded":{"url":"ftp://localhost/filepath/filename.txt","scheme":"ftp","netloc":"localhost","path":"/filepath/filename.txt","parent":"/filepath","filename":"filename.txt","stem":"filename","extension":"txt","path_list":["filepath","filename.txt"],"query":null,"fragment":null,"username":null,"password":null,"hostname":"localhost","port":null},"decoded":{"url":"ftp://localhost/filepath/filename.txt","scheme":"ftp","netloc":"localhost","path":"/filepath/filename.txt","parent":"/filepath","filename":"filename.txt","stem":"filename","extension":"txt","path_list":["filepath","filename.txt"],"query":null,"fragment":null,"username":null,"password":null,"hostname":"localhost","port":null}}''')
        self.assertEqual(jc.parsers.url.parse(data, quiet=True), expected)


    def test_url_http_ipv4(self):
        """
        Test HTTP URL with encodable characters (ipv4 host)
        """
        data = r'<URL:http://user{one:pass{two@127.0.0.1:8000/a space/b/c/d?q1=foo with {space}&q2=bar&q2=baz#frag{frag>'
        expected = json.loads(r'''{"url":"http://user{one:pass{two@127.0.0.1:8000/a space/b/c/d?q1=foo with {space}&q2=bar&q2=baz#frag{frag","scheme":"http","netloc":"user{one:pass{two@127.0.0.1:8000","path":"/a space/b/c/d","parent":"/a space/b/c","filename":"d","stem":"d","extension":null,"path_list":["a space","b","c","d"],"query":"q1=foo with {space}&q2=bar&q2=baz","query_obj":{"q1":["foo with {space}"],"q2":["bar","baz"]},"fragment":"frag{frag","username":"user{one","password":"pass{two","hostname":"127.0.0.1","port":8000,"encoded":{"url":"http://user%7Bone:pass%7Btwo@127.0.0.1:8000/a%20space/b/c/d?q1=foo+with+%7Bspace%7D&q2=bar&q2=baz#frag%7Bfrag","scheme":"http","netloc":"user%7Bone:pass%7Btwo@127.0.0.1:8000","path":"/a%20space/b/c/d","parent":"/a%20space/b/c","filename":"d","stem":"d","extension":null,"path_list":["a%20space","b","c","d"],"query":"q1=foo+with+%7Bspace%7D&q2=bar&q2=baz","fragment":"frag%7Bfrag","username":"user%7Bone","password":"pass%7Btwo","hostname":"127.0.0.1","port":8000},"decoded":{"url":"http://user{one:pass{two@127.0.0.1:8000/a space/b/c/d?q1=foo with {space}&q2=bar&q2=baz#frag{frag","scheme":"http","netloc":"user{one:pass{two@127.0.0.1:8000","path":"/a space/b/c/d","parent":"/a space/b/c","filename":"d","stem":"d","extension":null,"path_list":["a space","b","c","d"],"query":"q1=foo with {space}&q2=bar&q2=baz","fragment":"frag{frag","username":"user{one","password":"pass{two","hostname":"127.0.0.1","port":8000}}''')
        self.assertEqual(jc.parsers.url.parse(data, quiet=True), expected)


    def test_url_http_ipv6(self):
        """
        Test HTTP URL with encodable characters (ipv6 host)
        """
        data = r'<URL:http://user{one:pass{two@[1:2::127]:8000/a space/b/c/d?q1=foo with {space}&q2=bar&q2=baz#frag{frag>'
        expected = json.loads(r'''{"url":"http://user{one:pass{two@[1:2::127]:8000/a space/b/c/d?q1=foo with {space}&q2=bar&q2=baz#frag{frag","scheme":"http","netloc":"user{one:pass{two@[1:2::127]:8000","path":"/a space/b/c/d","parent":"/a space/b/c","filename":"d","stem":"d","extension":null,"path_list":["a space","b","c","d"],"query":"q1=foo with {space}&q2=bar&q2=baz","query_obj":{"q1":["foo with {space}"],"q2":["bar","baz"]},"fragment":"frag{frag","username":"user{one","password":"pass{two","hostname":"1:2::127","port":8000,"encoded":{"url":"http://user%7Bone:pass%7Btwo@[1:2::127]:8000/a%20space/b/c/d?q1=foo+with+%7Bspace%7D&q2=bar&q2=baz#frag%7Bfrag","scheme":"http","netloc":"user%7Bone:pass%7Btwo@[1:2::127]:8000","path":"/a%20space/b/c/d","parent":"/a%20space/b/c","filename":"d","stem":"d","extension":null,"path_list":["a%20space","b","c","d"],"query":"q1=foo+with+%7Bspace%7D&q2=bar&q2=baz","fragment":"frag%7Bfrag","username":"user%7Bone","password":"pass%7Btwo","hostname":"1:2::127","port":8000},"decoded":{"url":"http://user{one:pass{two@[1:2::127]:8000/a space/b/c/d?q1=foo with {space}&q2=bar&q2=baz#frag{frag","scheme":"http","netloc":"user{one:pass{two@[1:2::127]:8000","path":"/a space/b/c/d","parent":"/a space/b/c","filename":"d","stem":"d","extension":null,"path_list":["a space","b","c","d"],"query":"q1=foo with {space}&q2=bar&q2=baz","fragment":"frag{frag","username":"user{one","password":"pass{two","hostname":"1:2::127","port":8000}}''')
        self.assertEqual(jc.parsers.url.parse(data, quiet=True), expected)


    def test_url_http_domain(self):
        """
        Test HTTP URL with encodable characters (domain name host)
        """
        data = r'<URL:http://user{one:pass{two@www.example.com:8000/a space/b/c/d?q1=foo with {space}&q2=bar&q2=baz#frag{frag>'
        expected = json.loads(r'''{"url":"http://user{one:pass{two@www.example.com:8000/a space/b/c/d?q1=foo with {space}&q2=bar&q2=baz#frag{frag","scheme":"http","netloc":"user{one:pass{two@www.example.com:8000","path":"/a space/b/c/d","parent":"/a space/b/c","filename":"d","stem":"d","extension":null,"path_list":["a space","b","c","d"],"query":"q1=foo with {space}&q2=bar&q2=baz","query_obj":{"q1":["foo with {space}"],"q2":["bar","baz"]},"fragment":"frag{frag","username":"user{one","password":"pass{two","hostname":"www.example.com","port":8000,"encoded":{"url":"http://user%7Bone:pass%7Btwo@www.example.com:8000/a%20space/b/c/d?q1=foo+with+%7Bspace%7D&q2=bar&q2=baz#frag%7Bfrag","scheme":"http","netloc":"user%7Bone:pass%7Btwo@www.example.com:8000","path":"/a%20space/b/c/d","parent":"/a%20space/b/c","filename":"d","stem":"d","extension":null,"path_list":["a%20space","b","c","d"],"query":"q1=foo+with+%7Bspace%7D&q2=bar&q2=baz","fragment":"frag%7Bfrag","username":"user%7Bone","password":"pass%7Btwo","hostname":"www.example.com","port":8000},"decoded":{"url":"http://user{one:pass{two@www.example.com:8000/a space/b/c/d?q1=foo with {space}&q2=bar&q2=baz#frag{frag","scheme":"http","netloc":"user{one:pass{two@www.example.com:8000","path":"/a space/b/c/d","parent":"/a space/b/c","filename":"d","stem":"d","extension":null,"path_list":["a space","b","c","d"],"query":"q1=foo with {space}&q2=bar&q2=baz","fragment":"frag{frag","username":"user{one","password":"pass{two","hostname":"www.example.com","port":8000}}''')
        self.assertEqual(jc.parsers.url.parse(data, quiet=True), expected)


    def test_url_http_encoded(self):
        """
        Test HTTP URL with encoded characters
        """
        data = r'http://user%7Bone:pass%7Btwo@www.example.com:8000/a%20space/b/c/d?q1=foo+with+%7Bspace%7D&q2=bar&q2=baz#frag%7Bfrag'
        expected = json.loads(r'''{"url":"http://user%7Bone:pass%7Btwo@www.example.com:8000/a%20space/b/c/d?q1=foo+with+%7Bspace%7D&q2=bar&q2=baz#frag%7Bfrag","scheme":"http","netloc":"user%7Bone:pass%7Btwo@www.example.com:8000","path":"/a%20space/b/c/d","parent":"/a%20space/b/c","filename":"d","stem":"d","extension":null,"path_list":["a%20space","b","c","d"],"query":"q1=foo+with+%7Bspace%7D&q2=bar&q2=baz","query_obj":{"q1":["foo with {space}"],"q2":["bar","baz"]},"fragment":"frag%7Bfrag","username":"user%7Bone","password":"pass%7Btwo","hostname":"www.example.com","port":8000,"encoded":{"url":"http://user%257Bone:pass%257Btwo@www.example.com:8000/a%2520space/b/c/d?q1=foo+with+%257Bspace%257D&q2=bar&q2=baz#frag%257Bfrag","scheme":"http","netloc":"user%257Bone:pass%257Btwo@www.example.com:8000","path":"/a%2520space/b/c/d","parent":"/a%2520space/b/c","filename":"d","stem":"d","extension":null,"path_list":["a%2520space","b","c","d"],"query":"q1=foo+with+%257Bspace%257D&q2=bar&q2=baz","fragment":"frag%257Bfrag","username":"user%257Bone","password":"pass%257Btwo","hostname":"www.example.com","port":8000},"decoded":{"url":"http://user{one:pass{two@www.example.com:8000/a space/b/c/d?q1=foo with {space}&q2=bar&q2=baz#frag{frag","scheme":"http","netloc":"user{one:pass{two@www.example.com:8000","path":"/a space/b/c/d","parent":"/a space/b/c","filename":"d","stem":"d","extension":null,"path_list":["a space","b","c","d"],"query":"q1=foo with {space}&q2=bar&q2=baz","fragment":"frag{frag","username":"user{one","password":"pass{two","hostname":"www.example.com","port":8000}}''')
        self.assertEqual(jc.parsers.url.parse(data, quiet=True), expected)


    def test_url_http_encodable_host_and_port(self):
        """
        Test HTTP URL with encodable characters in the hostname and port
        """
        data = r'http://хост.домен:8080'
        expected = json.loads(r'''{"url":"http://хост.домен:8080","scheme":"http","netloc":"хост.домен:8080","path":null,"parent":null,"filename":null,"stem":null,"extension":null,"path_list":null,"query":null,"query_obj":null,"fragment":null,"username":null,"password":null,"hostname":"хост.домен","port":8080,"encoded":{"url":"http://%D1%85%D0%BE%D1%81%D1%82.%D0%B4%D0%BE%D0%BC%D0%B5%D0%BD:8080","scheme":"http","netloc":"%D1%85%D0%BE%D1%81%D1%82.%D0%B4%D0%BE%D0%BC%D0%B5%D0%BD:8080","path":null,"parent":null,"filename":null,"stem":null,"extension":null,"path_list":null,"query":null,"fragment":null,"username":null,"password":null,"hostname":"%D1%85%D0%BE%D1%81%D1%82.%D0%B4%D0%BE%D0%BC%D0%B5%D0%BD","port":8080},"decoded":{"url":"http://хост.домен:8080","scheme":"http","netloc":"хост.домен:8080","path":null,"parent":null,"filename":null,"stem":null,"extension":null,"path_list":null,"query":null,"fragment":null,"username":null,"password":null,"hostname":"хост.домен","port":8080}}''')
        self.assertEqual(jc.parsers.url.parse(data, quiet=True), expected)


    def test_url_http_encoded_host_and_port(self):
        """
        Test HTTP URL with encoded characters in the hostname and port
        """
        data = r'http://%D1%85%D0%BE%D1%81%D1%82.%D0%B4%D0%BE%D0%BC%D0%B5%D0%BD:%38%38'
        expected = json.loads(r'''{"url":"http://%D1%85%D0%BE%D1%81%D1%82.%D0%B4%D0%BE%D0%BC%D0%B5%D0%BD:%38%38","scheme":"http","netloc":"%D1%85%D0%BE%D1%81%D1%82.%D0%B4%D0%BE%D0%BC%D0%B5%D0%BD:%38%38","path":null,"parent":null,"filename":null,"stem":null,"extension":null,"path_list":null,"query":null,"query_obj":null,"fragment":null,"username":null,"password":null,"hostname":"%D1%85%D0%BE%D1%81%D1%82.%D0%B4%D0%BE%D0%BC%D0%B5%D0%BD","port":88,"encoded":{"url":"http://%25D1%2585%25D0%25BE%25D1%2581%25D1%2582.%25D0%25B4%25D0%25BE%25D0%25BC%25D0%25B5%25D0%25BD:%2538%2538","scheme":"http","netloc":"%25D1%2585%25D0%25BE%25D1%2581%25D1%2582.%25D0%25B4%25D0%25BE%25D0%25BC%25D0%25B5%25D0%25BD:%2538%2538","path":null,"parent":null,"filename":null,"stem":null,"extension":null,"path_list":null,"query":null,"fragment":null,"username":null,"password":null,"hostname":"%25D1%2585%25D0%25BE%25D1%2581%25D1%2582.%25D0%25B4%25D0%25BE%25D0%25BC%25D0%25B5%25D0%25BD","port":88},"decoded":{"url":"http://хост.домен:88","scheme":"http","netloc":"хост.домен:88","path":null,"parent":null,"filename":null,"stem":null,"extension":null,"path_list":null,"query":null,"fragment":null,"username":null,"password":null,"hostname":"хост.домен","port":88}}''')
        self.assertEqual(jc.parsers.url.parse(data, quiet=True), expected)


    def test_url_http_encoded_host_and_invalid_port(self):
        """
        Test HTTP URL with encoded characters in the hostname and an invalid encoded port
        """
        data = r'http://хост.домен:%38{%38#frag'
        expected = json.loads(r'''{"url":"http://хост.домен:%38{%38#frag","scheme":"http","netloc":"хост.домен:%38{%38","path":null,"parent":null,"filename":null,"stem":null,"extension":null,"path_list":null,"query":null,"query_obj":null,"fragment":"frag","username":null,"password":null,"hostname":"хост.домен","port":null,"encoded":{"url":"http://%D1%85%D0%BE%D1%81%D1%82.%D0%B4%D0%BE%D0%BC%D0%B5%D0%BD:%2538%7B%2538#frag","scheme":"http","netloc":"%D1%85%D0%BE%D1%81%D1%82.%D0%B4%D0%BE%D0%BC%D0%B5%D0%BD:%2538%7B%2538","path":null,"parent":null,"filename":null,"stem":null,"extension":null,"path_list":null,"query":null,"fragment":"frag","username":null,"password":null,"hostname":"%D1%85%D0%BE%D1%81%D1%82.%D0%B4%D0%BE%D0%BC%D0%B5%D0%BD","port":null},"decoded":{"url":"http://хост.домен:8{8#frag","scheme":"http","netloc":"хост.домен:8{8","path":null,"parent":null,"filename":null,"stem":null,"extension":null,"path_list":null,"query":null,"fragment":"frag","username":null,"password":null,"hostname":"хост.домен","port":null}}''')
        self.assertEqual(jc.parsers.url.parse(data, quiet=True), expected)


    def test_url_http_path_ends_in_slash(self):
        """
        Test HTTP URL with a forward slash as the last part of the path
        """
        data = r'https://pypi.org/project/jc/1.20.3/'
        expected = json.loads(r'''{"url":"https://pypi.org/project/jc/1.20.3/","scheme":"https","netloc":"pypi.org","path":"/project/jc/1.20.3/","parent":"/project/jc","filename":"1.20.3","stem":"1.20","extension":"3","path_list":["project","jc","1.20.3"],"query":null,"query_obj":null,"fragment":null,"username":null,"password":null,"hostname":"pypi.org","port":null,"encoded":{"url":"https://pypi.org/project/jc/1.20.3/","scheme":"https","netloc":"pypi.org","path":"/project/jc/1.20.3/","parent":"/project/jc","filename":"1.20.3","stem":"1.20","extension":"3","path_list":["project","jc","1.20.3"],"query":null,"fragment":null,"username":null,"password":null,"hostname":"pypi.org","port":null},"decoded":{"url":"https://pypi.org/project/jc/1.20.3/","scheme":"https","netloc":"pypi.org","path":"/project/jc/1.20.3/","parent":"/project/jc","filename":"1.20.3","stem":"1.20","extension":"3","path_list":["project","jc","1.20.3"],"query":null,"fragment":null,"username":null,"password":null,"hostname":"pypi.org","port":null}}''')
        self.assertEqual(jc.parsers.url.parse(data, quiet=True), expected)


    def test_url_http_path_only_slash(self):
        """
        Test HTTP URL with a forward slash as the last only part of the path
        """
        data = r'https://pypi.org/'
        expected = json.loads(r'''{"url":"https://pypi.org/","scheme":"https","netloc":"pypi.org","path":"/","parent":"/","filename":null,"stem":null,"extension":null,"path_list":null,"query":null,"query_obj":null,"fragment":null,"username":null,"password":null,"hostname":"pypi.org","port":null,"encoded":{"url":"https://pypi.org/","scheme":"https","netloc":"pypi.org","path":"/","parent":"/","filename":null,"stem":null,"extension":null,"path_list":null,"query":null,"fragment":null,"username":null,"password":null,"hostname":"pypi.org","port":null},"decoded":{"url":"https://pypi.org/","scheme":"https","netloc":"pypi.org","path":"/","parent":"/","filename":null,"stem":null,"extension":null,"path_list":null,"query":null,"fragment":null,"username":null,"password":null,"hostname":"pypi.org","port":null}}''')
        self.assertEqual(jc.parsers.url.parse(data, quiet=True), expected)


    def test_url_http_path_no_end_slash(self):
        """
        Test HTTP URL with no forward slash at the end
        """
        data = r'https://pypi.org'
        expected = json.loads(r'''{"url":"https://pypi.org","scheme":"https","netloc":"pypi.org","path":null,"parent":null,"filename":null,"stem":null,"extension":null,"path_list":null,"query":null,"query_obj":null,"fragment":null,"username":null,"password":null,"hostname":"pypi.org","port":null,"encoded":{"url":"https://pypi.org","scheme":"https","netloc":"pypi.org","path":null,"parent":null,"filename":null,"stem":null,"extension":null,"path_list":null,"query":null,"fragment":null,"username":null,"password":null,"hostname":"pypi.org","port":null},"decoded":{"url":"https://pypi.org","scheme":"https","netloc":"pypi.org","path":null,"parent":null,"filename":null,"stem":null,"extension":null,"path_list":null,"query":null,"fragment":null,"username":null,"password":null,"hostname":"pypi.org","port":null}}''')
        self.assertEqual(jc.parsers.url.parse(data, quiet=True), expected)


    def test_url_file_path(self):
        """
        Test a normal file path URL
        """
        data = r'/this/is my/path/to the file/myfile.tar.gz'
        expected = json.loads(r'''{"url":"/this/is my/path/to the file/myfile.tar.gz","scheme":null,"netloc":null,"path":"/this/is my/path/to the file/myfile.tar.gz","parent":"/this/is my/path/to the file","filename":"myfile.tar.gz","stem":"myfile.tar","extension":"gz","path_list":["this","is my","path","to the file","myfile.tar.gz"],"query":null,"query_obj":null,"fragment":null,"username":null,"password":null,"hostname":null,"port":null,"encoded":{"url":"/this/is%20my/path/to%20the%20file/myfile.tar.gz","scheme":null,"netloc":null,"path":"/this/is%20my/path/to%20the%20file/myfile.tar.gz","parent":"/this/is%20my/path/to%20the%20file","filename":"myfile.tar.gz","stem":"myfile.tar","extension":"gz","path_list":["this","is%20my","path","to%20the%20file","myfile.tar.gz"],"query":null,"fragment":null,"username":null,"password":null,"hostname":null,"port":null},"decoded":{"url":"/this/is my/path/to the file/myfile.tar.gz","scheme":null,"netloc":null,"path":"/this/is my/path/to the file/myfile.tar.gz","parent":"/this/is my/path/to the file","filename":"myfile.tar.gz","stem":"myfile.tar","extension":"gz","path_list":["this","is my","path","to the file","myfile.tar.gz"],"query":null,"fragment":null,"username":null,"password":null,"hostname":null,"port":null}}''')
        self.assertEqual(jc.parsers.url.parse(data, quiet=True), expected)


if __name__ == '__main__':
    unittest.main()



import os
import unittest
import json
import jc.parsers.xml

# fix for whether tests are run directly or via runtests.sh
try:
    from ._vendor.packaging import version
except:
    from _vendor.packaging import version  # type: ignore

# check the version of installed xmltodict library
try:
    import xmltodict
    XMLTODICT_INSTALLED = True
    XMLTODICT_0_13_0_OR_HIGHER = version.parse(xmltodict.__version__) >= version.parse('0.13.0')  # type: ignore
except:
    XMLTODICT_INSTALLED = False


THIS_DIR = os.path.dirname(os.path.abspath(__file__))

@unittest.skipIf(not XMLTODICT_INSTALLED, 'xmltodict library not installed')
class MyTests(unittest.TestCase):

    # input
    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/xml-cd_catalog.xml'), 'r', encoding='utf-8') as f:
        generic_xml_cd_catalog = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/xml-foodmenu.xml'), 'r', encoding='utf-8') as f:
        generic_xml_foodmenu = f.read()

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/xml-nmap.xml'), 'r', encoding='utf-8') as f:
        generic_xml_nmap = f.read()

    # output
    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/xml-cd_catalog.json'), 'r', encoding='utf-8') as f:
        generic_xml_cd_catalog_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/xml-foodmenu.json'), 'r', encoding='utf-8') as f:
        generic_xml_foodmenu_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/xml-nmap.json'), 'r', encoding='utf-8') as f:
        generic_xml_nmap_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/xml-nmap-nocomment.json'), 'r', encoding='utf-8') as f:
        generic_xml_nmap_nocomment_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/xml-nmap-raw.json'), 'r', encoding='utf-8') as f:
        generic_xml_nmap_r_json = json.loads(f.read())

    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/xml-nmap-raw-nocomment.json'), 'r', encoding='utf-8') as f:
        generic_xml_nmap_r_nocomment_json = json.loads(f.read())


    def test_xml_nodata(self):
        """
        Test xml parser with no data
        """
        self.assertEqual(jc.parsers.xml.parse('', quiet=True), [])

    def test_xml_nodata_r(self):
        """
        Test xml parser with no data and raw output
        """
        self.assertEqual(jc.parsers.xml.parse('', raw=True, quiet=True), [])

    def test_xml_cd_catalog(self):
        """
        Test the cd catalog xml file
        """
        self.assertEqual(jc.parsers.xml.parse(self.generic_xml_cd_catalog, quiet=True), self.generic_xml_cd_catalog_json)

    def test_xml_foodmenu(self):
        """
        Test the food menu xml file
        """
        self.assertEqual(jc.parsers.xml.parse(self.generic_xml_foodmenu, quiet=True), self.generic_xml_foodmenu_json)

    def test_xml_nmap(self):
        """
        Test nmap xml output
        """
        if XMLTODICT_0_13_0_OR_HIGHER:
            self.assertEqual(jc.parsers.xml.parse(self.generic_xml_nmap, quiet=True), self.generic_xml_nmap_json)
        else:
            print('\n### Using older version of xmltodict library. Testing without comment support. ### ... ')
            self.assertEqual(jc.parsers.xml.parse(self.generic_xml_nmap, quiet=True), self.generic_xml_nmap_nocomment_json)

    def test_xml_nmap_r(self):
        """
        Test nmap xml raw output
        """
        if XMLTODICT_0_13_0_OR_HIGHER:
            self.assertEqual(jc.parsers.xml.parse(self.generic_xml_nmap, raw=True, quiet=True), self.generic_xml_nmap_r_json)
        else:
            print('\n### Using older version of xmltodict library. Testing without comment support. ### ... ')
            self.assertEqual(jc.parsers.xml.parse(self.generic_xml_nmap, raw=True, quiet=True), self.generic_xml_nmap_r_nocomment_json)


if __name__ == '__main__':
    unittest.main()



import os
import json
import unittest
import jc.parsers.syslog_s

THIS_DIR = os.path.dirname(os.path.abspath(__file__))

# To create streaming output use:
# $ cat syslog.out | jc --syslog-s | jello -c > syslog-streaming.json


class MyTests(unittest.TestCase):

    # input
    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/syslog-5424.out'), 'r', encoding='utf-8') as f:
        syslog = f.read()

    # output
    with open(os.path.join(THIS_DIR, os.pardir, 'tests/fixtures/generic/syslog-5424-streaming.json'), 'r', encoding='utf-8') as f:
        syslog_streaming_json = json.loads(f.read())


    def test_syslog_s_nodata(self):
        """
        Test 'syslog' with no data
        """
        self.assertEqual(list(jc.parsers.syslog_s.parse([], quiet=True)), [])

    def test_syslog_s(self):
        """
        Test syslog file
        """
        self.assertEqual(list(jc.parsers.syslog_s.parse(self.syslog.splitlines(), quiet=True)), self.syslog_streaming_json)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# Generate README.md from jc metadata using jinja2 templates
import jc.cli
import jc.lib
from jinja2 import Environment, FileSystemLoader

file_loader = FileSystemLoader('templates')
env = Environment(loader=file_loader)
template = env.get_template('readme_template')
output = template.render(parsers=jc.lib.all_parser_info(),
                         info=jc.cli.JcCli.about_jc())

with open('README.md', 'w') as f:
    f.write(output)



import os
import unittest
import json
from typing import Dict
from jc.parsers.zpool_status import parse

THIS_DIR = os.path.dirname(os.path.abspath(__file__))


class MyTests(unittest.TestCase):
    f_in: Dict = {}
    f_json: Dict = {}

    @classmethod
    def setUpClass(cls):
        fixtures = {
            'zpool_status': (
                'fixtures/generic/zpool-status-v.out',
                'fixtures/generic/zpool-status-v.json'),
            'zpool_status2': (
                'fixtures/generic/zpool-status-v2.out',
                'fixtures/generic/zpool-status-v2.json'),
            'zpool_status3': (
                'fixtures/generic/zpool-status-v3.out',
                'fixtures/generic/zpool-status-v3.json'),
            'zpool_status4': (
                'fixtures/generic/zpool-status-v4.out',
                'fixtures/generic/zpool-status-v4.json'),
            'zpool_status_spares': (
                'fixtures/generic/zpool-status-spares.out',
                'fixtures/generic/zpool-status-spares.json')
        }

        for file, filepaths in fixtures.items():
            with open(os.path.join(THIS_DIR, filepaths[0]), 'r', encoding='utf-8') as a, \
                 open(os.path.join(THIS_DIR, filepaths[1]), 'r', encoding='utf-8') as b:
                cls.f_in[file] = a.read()
                cls.f_json[file] = json.loads(b.read())


    def test_zpool_status_nodata(self):
        """
        Test 'zpool_status' with no data
        """
        self.assertEqual(parse('', quiet=True), [])


    def test_zpool_status_v(self):
        """
        Test 'zpool status -v'
        """
        self.assertEqual(
            parse(self.f_in['zpool_status'], quiet=True),
            self.f_json['zpool_status']
        )

    def test_zpool_status_v_2(self):
        """
        Test 'zpool status -v' #2
        """
        self.assertEqual(
            parse(self.f_in['zpool_status2'], quiet=True),
            self.f_json['zpool_status2']
        )

    def test_zpool_status_v_3(self):
        """
        Test 'zpool status -v' #3
        """
        self.assertEqual(
            parse(self.f_in['zpool_status3'], quiet=True),
            self.f_json['zpool_status3']
        )

    def test_zpool_status_v_with_tabs(self):
        """
        Test 'zpool status -v' with tabs instead of spaces
        """
        self.assertEqual(
            parse(self.f_in['zpool_status4'], quiet=True),
            self.f_json['zpool_status4']
        )

    def test_zpool_status_spares(self):
        """
        Test 'zpool status' with spares section
        """
        self.assertEqual(
            parse(self.f_in['zpool_status_spares'], quiet=True),
            self.f_json['zpool_status_spares']
        )


if __name__ == '__main__':
    unittest.main()
